# Grammatical text structure.
# Note: the input gets executed as python code (see buildMethod),
# so the grammar should be built from source in the system.
from types import ClassType as newClassObject


def buildCode(name, parameters = None, code = None, docstring = None):
    if not code:
        code = 'pass'
    elif isinstance(code, list):
        code = nls(code)

    def build():
        yield 'def %s(%s):' % (name, ', '.join(parameters or []))

        if docstring:
            yield indent(repr(docstring))

        yield indent(code)

    return nls(build()) + '\n'

def buildMethod(code, name, globals = None):
    if globals is None:
        from builtins import globals
        globals = globals()

    # print code

    ns = {}
    exec(code, globals, ns)
    return ns[name]

def integrateMethod(values, name, parameters, code, docstring, globals = None):
    values[name] = buildMethod(buildCode(name, parameters = parameters,
                                         code = code, docstring = docstring),
                               name, globals)

def integrateMethodMap(values, name, mapping, globals = None):
    return integrateMethod(values, name,
                           mapping.get('parameters'),
                           mapping.get('code'),
                           mapping.get('documentation'),
                           globals)

def _baseClasses(c, default):
    if not c:
        c = default
    if not isinstance(c, (list, tuple)):
        return (c,)

    return c


class Lexicon:
    @classmethod
    def Build(self, className, lexicon, baseClass = None, globals = None):
        # Build lexer input class.
        values = dict()

        # Mandate tokens.

        # Does absence in these names cause this error?
        #   File "C:\Python27\lib\site-packages\ply\lex.py", line 363, in token
        #     func.__name__, newtok.type), lexdata[lexpos:])
        # LexError: <string>:1: Rule 't_NAME' returned an unknown token type 'if'

        tokens = lexicon.get('names')
        if isinstance(tokens, str):
            tokens = tokens.split()
        if not tokens:
            tokens = []

        values['tokens'] = tokens

        # Optional instance constructor.
        try: init = lexicon['init']
        except KeyError: pass
        else: integrateMethodMap(values, '__init__', init, globals)

        # Token match methods.
        for (name, tokdef) in lexicon.items():
            if name.startswith('t_'):
                if isinstance(tokdef, str):
                    values[name] = tokdef

                    # No, this produces lexfuncs that return None!
                    # integrateMethod(values, name,
                    #                 ['self', 't'], None,
                    #                 docstring = tokdef,
                    #                 globals = globals)

                else:
                    integrateMethod(values, name,
                                    tokdef.get('parameters', ['self', 't']),
                                    tokdef.get('match'),
                                    tokdef.get('pattern'),
                                    globals = globals)

            elif isinstance(tokdef, dict):
                integrateMethodMap(values, name, tokdef,
                                   globals = globals)


        baseClass = _baseClasses(baseClass, self)
        return newClassObject(className, baseClass, values)

    def __init__(self, grammar):
        self.grammar = grammar


class Rules:
    @classmethod
    def Build(self, className, rules, baseClass = None, globals = None):
        # Build rule/production input class.
        values = dict()

        # Optional instance constructor.
        try: init = rules['init']
        except KeyError: pass
        else: integrateMethodMap(values, '__init__', init, globals)

        # Rule production methods.
        for (name, ruledef) in rules.items():
            if name.startswith('p_'):
                if isinstance(ruledef, str):
                    integrateMethod(values, name, ['self', 'p'],
                                    None, ruledef, globals)
                else:
                    integrateMethod(values, name,
                                    ruledef.get('parameters', ['self', 'p']),
                                    ruledef.get('production', 'p[0] = p[1]'),
                                    ruledef.get('pattern'),
                                    globals)

            elif isinstance(ruledef, dict):
                integrateMethodMap(values, name, ruledef, globals)


        baseClass = _baseClasses(baseClass, self)
        return newClassObject(className, baseClass, values)

    def __init__(self, grammar):
        self.grammar = grammar
        self.tokens = grammar.lexicon.tokens
        if hasattr(grammar.lexicon, 'literal'):
            self.literals = grammar.lexicon.literals


class Grammar:
    @classmethod
    def FromStructure(self, className, value, **kwd):
        # Note: this produces a Grammar subclass with defined lexicon/rules classes.
        # The grammar class must be instantiated with the starting rule: if you specify
        # the 'start' key, it will be.

        globals = kwd.get('globals')

        lexicon = value['lexicon']
        lexiconClass = kwd.get('lexiconClass')
        lexicon = Lexicon.Build('lexicon', lexicon, lexiconClass, globals = globals)

        rules = value['rules']
        ruleClass = kwd.get('ruleClass')
        rules = Rules.Build('rules', rules, ruleClass, globals = globals)

        values = dict(lexiconClass = lexicon, ruleClass = rules)

        for (name, method) in value.items():
            if name not in ['lexicon', 'rules', 'start']:
                if isinstance(method, dict):
                    integrateMethodMap(values, name, method, globals)


        baseClass = _baseClasses(kwd.get('grammarClass'), self)
        grammarClass = newClassObject(className, baseClass, values)

        try: start = value['start']
        except KeyError:
            return grammarClass

        return grammarClass(start, value.get('debug'))


    def __init__(self, start, debug = False):
        from ply import lex, yacc
        self.start = start

        self.lexicon = self.lexiconClass(self)
        self.rules = self.ruleClass(self)

        # try: from ph.lang.layer import gs_indent_lexer
        try: from stuphos.language.gen import gs_indent_lexer
        except ImportError:
            # todo: optionally mandate this.
            self.lexer = lex.lex(object = self.lexicon)
        else:
            self.lexer = gs_indent_lexer.IndentLexer(self.lexicon)

        if debug:
            import sys
            debug = yacc.PlyLogger(sys.stderr)

        ERRORLOG = False
        self.debug = debug
        self.parser = yacc.yacc(module = self.rules,
                                debug = debug,
                                errorlog = None, #  if ERRORLOG else yacc.NullLogger(),
                                tabmodule = 'parsetab_%s' % start,
                                debugfile = 'parser_%s.out' % start,
                                outputdir = '.',
                                start = start,
                                check_recursion = False)

        # Other yacc options:
        #   debug = 0   ...   No debugging output, including writing parser.out
        #   outputdir = "x"   ...   Location of written tab module.
        #   method = "SLR"   ...   Use simpler parser for faster generation.
        #   errorlog = yacc.NullLogger()

    def parse(self, source):
        self.reset_errors()
        try: return self.parser.parse(source, lexer = self.lexer, debug = self.debug) # , tracking = 1)
        finally:
            self.report_errors()

    def tokenize(self, source):
        self.lexer.input(source)
        return self.lexer.token_stream

        # try:
        #     while True:
        #         yield self.lexer.next()

        # except StopIteration:
        #     pass

    def report_errors(self):
        try:
            for e in self.lexical_errors:
                print('Illegal character: %r!' % e)
            for e in self.parse_errors:
                print('Syntax error: %r!' % e)
        except IOError:
            # This might happen if stdout gets closed!
            pass

        self.reset_errors()

    def reset_errors(self):
        self.lexical_errors = []
        self.parse_errors = []


Rules.base = Rules
Lexicon.base = Lexicon
Grammar.base = Grammar


girl = '''
server:
    # start: program
    start: wmc_document
    debug: true

    lexicon:
        names::
            ASSIGN         AT             BANG           COLON          
            COMMA          DECIMAL        DECIMALSTRING  DEDENT         
            DOT            ELLIPSE        ENDMARKER      EQUAL          
            GREATERTHAN    GREATERTHANEQ  INDENT         INTEGER        
            LBRACKET       LESSTHAN       LESSTHANEQ     LPAREN         
            MINUS          MINUSASSIGN    MODASSIGN      MODULO         
            NAME           NEWLINE        NOT            NOTEQUAL       
            NUMBER         PLUS           PLUSASSIGN     RBRACKET       
            RPAREN         SEMICOLON      SLASH          SLASHASSIGN    
            STAR           STARASSIGN     STRING         WS             
            IF             DEF            ELIF           ELSE
            WHILE          TRY            EXCEPT         FINALLY
            FOR            IN             NOT            RETURN

            DOLLAR TRUE FALSE NULL LBRACE RBRACE WMCLISTVALUE WMCMAPVALUE WMCHERESTRING


        # Simple expressional tokens:
        t_PLUS: '\+'
        t_MINUS: '-'
        t_STAR: '\*'
        t_SLASH: '/'
        t_LPAREN: '\('
        t_RPAREN: '\)'
        t_LBRACKET: '\['
        t_RBRACKET: '\]'

        t_BANG: '!'
        t_NOTEQUAL: '!='
        t_AT: '@'
        t_MODULO: '%'

        t_EQUAL: '=='
        t_LESSTHAN: '<'
        t_LESSTHANEQ: '<='
        t_COMMA: ','
        t_DOT: '\.'
        t_GREATERTHAN: '>'
        t_GREATERTHANEQ: '>='

        t_NOT: 'not'

        t_ELLIPSE: '\.\.\.'

        t_COLON: '\:'
        t_SEMICOLON: ';'
        t_ASSIGN: '='
        t_PLUSASSIGN: '\+='
        t_MINUSASSIGN: '-='
        t_STARASSIGN: '\*='
        t_SLASHASSIGN: '/='
        t_MODASSIGN: '%='

        t_DOLLAR: '\\$'
        t_TRUE: 'true'
        t_FALSE: 'false'
        t_NULL: 'null'

        t_WMCLISTVALUE: '0'
        t_WMCMAPVALUE: '0'
        t_WMCHERESTRING: '0'


        # Compilable rules:

        # Apparently this wants to match trailing colons.......?
        # t_DECIMAL:
        #     # ? Is this parsing integers??  probably because of unordered tokpat?
        #     pattern: '((\d+.\d*)|(\d*.\d+))'
        #     match::
        #         t.value = float(t.value)
        #         return t

        t_NUMBER:
            pattern: '(\d+(\.\d*)?|\.\d+)([eE][-+]? \d+)?'
            match::
                v = t.value
                if v.isdigit():
                    v = int(v)
                else:
                    v = decimal.Decimal(v)

                t.value = v
                return t

        # t_STRING:
        #     # pattern: "'([^\\']+|\\'|\\\\|\\r|\\n)*'"
        #     pattern: 's'
        #     match::
        #         t.value=t.value[1:-1].decode("string-escape")
        #         return t

        init:
            parameters: [self, grammar]
            code::
                self.base.__init__(self, grammar)
                self.reserved = ['while', 'if']

        t_NAME:
            pattern: '[\$a-zA-Z_][\$a-zA-Z_0-9]*'
            match::
                # Check for reserved words.
                if t.value in self.reserved:
                    # t.type = t.value # must be in lexer.tokens
                    t.reserved = True

                # t.type = self.reserved.get(t.value, 'NAME')

                # This is to demonstrate that the token value can acquire a complex
                # value, although since yacc only looks at the 'value' attribute, the
                # token should set any _other_ attributes for further processing...
                # t.value = (t.value, symbol_lookup(t.value))

                return t

        t_comment:
            pattern: "[ ]*\\043[^\\n]*"

        t_WS:
            pattern: '[ ]+'
            match::
                # Note: This is not used in the parser, but by the gs_indent_lexer layer.
                # This terminal will generate a warning, but it is required.
                if t.lexer.at_line_start and t.lexer.paren_count == 0:
                    return t

        t_newline:
            pattern: '\\n+'
            match::
                t.lexer.lineno += len(t.value)
                t.type = 'NEWLINE'
                return t

        t_error:
            match::
                self.grammar.lexical_errors.append(t.value[0])
                t.lexer.skip(1)

    rules:
        p_error:
            production::
                self.grammar.parse_errors.append(p)


        p_testlist:
            pattern::
                testlist : testlist_multi COMMA
                         | testlist_multi

            production::
                if len(p) == 2:
                    p[0] = p[1]
                else:
                    # May need to promote singleton to tuple
                    if isinstance(p[1], list):
                        p[0] = p[1]
                    else:
                        p[0] = [p[1]]

                # Convert into a list?
                if isinstance(p[0], list):
                    p[0] = List(p[0], self.grammar.lexer.lineno)


        p_arglist::
            arglist : arglist COMMA argument
                               | argument

        p_argument::
            argument : test

        p_atom_list::
                    atom : LPAREN testlist RPAREN
                         | LBRACKET testlist RBRACKET
                         | LBRACKET RBRACKET


        p_atom_name::
            atom : NAME

        p_atom_number::
                    atom : NUMBER
                         | STRING


        p_comparison::
                    comparison : comparison PLUS comparison
                               | comparison MINUS comparison
                               | comparison STAR comparison
                               | comparison SLASH comparison
                               | comparison MODULO comparison
                               | comparison NOTEQUAL comparison
                               | comparison EQUAL comparison
                               | comparison LESSTHAN comparison
                               | comparison LESSTHANEQ comparison
                               | comparison GREATERTHAN comparison
                               | comparison GREATERTHANEQ comparison
                               | PLUS comparison
                               | MINUS comparison
                               | NOT comparison
                               | power


        p_compound_stmt::
                    compound_stmt : if_stmt
                                  | while_loop
                                  | for_loop
                                  | try_except
                                  | funcdef


        p_else::
            else : ELSE COLON suite

        p_elseif::
            elseif : ELIF test COLON suite

        p_elseif_chain::
                    elseif_chain : elseif_chain elseif
                                 | elseif


        p_error::


        p_except::
                    except : EXCEPT LPAREN except_type_list RPAREN COMMA NAME COLON suite
                           | EXCEPT NAME NAME COLON suite


        p_except_chain::
                    except_chain : except_chain except
                                 | except


        p_except_type_list::
                    except_type_list : except_type_list COMMA NAME
                                     | NAME


        p_expr_stmt::
                    expr_stmt : lvalue ASSIGN testlist
                              | lvalue PLUSASSIGN testlist
                              | lvalue MINUSASSIGN testlist
                              | lvalue STARASSIGN testlist
                              | lvalue SLASHASSIGN testlist
                              | lvalue MODASSIGN testlist
                              | testlist


        p_flow_stmt::
            flow_stmt : return_stmt

        p_for_loop::
            for_loop : FOR NAME IN test COLON suite

        p_funcdef::
            funcdef : DEF NAME parameters COLON suite

        p_if_stmt::

                    if_stmt : IF test COLON suite elseif_chain else
                            | IF test COLON suite elseif_chain
                            | IF test COLON suite else
                            | IF test COLON suite


        p_lvalue_testlist::
            lvalue : testlist

        p_module::
            module : program ENDMARKER

        p_parameters::
            parameters : LPAREN RPAREN
                                  | LPAREN varargslist RPAREN

        p_power::
                    power : atom
                          | power DOT NAME
                          | power LBRACKET testlist RBRACKET
                          | power trailer


        p_program::
                    program : program NEWLINE
                            | program stmt
                            | NEWLINE
                            | stmt


        p_return_stmt::
                    return_stmt : RETURN
                                | RETURN testlist


        p_simple_stmt::
                    simple_stmt : small_stmts NEWLINE
                                | small_stmts SEMICOLON NEWLINE


        p_small_stmt::
                    small_stmt : flow_stmt
                               | expr_stmt
                               | wmc


        p_small_stmts::
                    small_stmts : small_stmts SEMICOLON small_stmt
                                | small_stmt


        p_stmt_compound::
            stmt : compound_stmt

        p_stmt_simple::
            stmt : simple_stmt

        p_stmts::
                    stmts : stmts stmt
                          | stmt


        p_suite::
                    suite : simple_stmt
                          | NEWLINE INDENT stmts DEDENT


        p_test::
            test : comparison

        p_testlist::
                    testlist : testlist_multi COMMA
                             | testlist_multi


        p_testlist_multi::
                    testlist_multi : testlist_multi COMMA test
                                   | test


        p_trailer::
                    trailer : LPAREN RPAREN
                            | LPAREN arglist RPAREN


        p_try_except::
                    try_except : TRY COLON suite except_chain EXCEPT COLON suite FINALLY COLON suite
                               | TRY COLON suite except_chain EXCEPT COLON suite
                               | TRY COLON suite except_chain FINALLY COLON suite
                               | TRY COLON suite EXCEPT COLON suite
                               | TRY COLON suite FINALLY COLON suite


        p_varargslist::
            varargslist : varargslist COMMA NAME
                                   | NAME

        p_while_loop::
            while_loop : WHILE test COLON suite

        init:
            parameters: [self, grammar]
            code::
                self.base.__init__(self, grammar)

                class WMC:
                    def __init__(self, name, package = None, subtype = None, value = None):
                        self.name = name
                        self.package = package
                        self.subtype = subtype
                        self.value = value

                    @property
                    def defn(self):
                        return '%s(%s$%s)' % (self.name, self.package, self.subtype)

                    def __repr__(self):
                        return self.defn

                    @nling
                    def __str__(self):
                        yield '%s:' % self.defn
                        yield indent(str(self.value))


                self.WMC = WMC

        # YAML
        p_wmc_document:
            pattern::
                wmc_document : wmc ENDMARKER

        # wmc_stmt
        p_wmc:
            pattern::
                wmc : NAME LPAREN NAME DOLLAR NAME RPAREN COLON wmcValue
                    | NAME LPAREN DOLLAR NAME RPAREN COLON wmcValue
                    | NAME LPAREN NAME RPAREN COLON wmcValue
                    | NAME COLON wmcValue
                    | wmcList

            production::
                if len(p) == 9:
                    p[0] = self.WMC(p[1], package = p[3], subtype = p[5], value = p[8])
                elif len(p) == 8:
                    p[0] = self.WMC(p[1], subtype = p[4], value = p[7])
                elif len(p) == 7:
                    p[0] = self.WMC(p[1], subtype = p[3], value = p[6])
                elif len(p) == 4:
                    p[0] = self.WMC(p[1], value = p[3])
                else:
                    # XXX Absorbs SyntaxErrors: yacc.py: 510
                    raise RuntimeError('wmc production length = %d' % len(p))

        p_wmcList:
            pattern::
                wmcList : wmcValue

            # : wmc wmcList
            # | wmc

        p_wmcValue:
            # | NEWLINE INDENT wmc NEWLINE DEDENT

            pattern::
                wmcValue : wmcSingleValue NEWLINE
                         | COLON NEWLINE INDENT WMCHERESTRING DEDENT NEWLINE
                         | NEWLINE INDENT wmcListValue NEWLINE DEDENT
                         | NEWLINE INDENT wmc DEDENT

            # replaced by wmcValue : wmcSingleValue NEWLINE
            # | NEWLINE INDENT wmcSingleValue NEWLINE

            production::
                if len(p) == 7:
                    p[0] = p[4]
                elif len(p) in [5, 6]:
                    p[0] = p[3]
                else:
                    p[0] = p[1]

        p_wmcSingleValue:
            # Need 'rest of line'
            # todo: get rid of NAME
            pattern::
                wmcSingleValue : STRING
                               | INTEGER
                               | DECIMAL
                               | NUMBER
                               | NAME
                               | TRUE
                               | FALSE
                               | NULL
                               | LBRACKET WMCLISTVALUE RBRACKET
                               | LBRACE WMCMAPVALUE RBRACE

            production::
                if len(p) == 4:
                    p[0] = p[2]
                else:
                    p[0] = p[1]

        p_wmcListValue:
            pattern::
                wmcListValue : NEWLINE INDENT MINUS wmcSingleValue wmcListValue NEWLINE

            production::
                p[0] = [p[4]] + p[5]

'''

g2 = '''
if 1:
    server:
        start: wmc_document
        debug: true

        lexicon:
            names:
                ASSIGN

            t_PLUS: name
            t_MINUS: name
            t_STAR: name

            t_DECIMAL:
                pattern: name
                match:
                    name

while x[y(z) + t]:
    if n(z):
        while true:
            ok:
                this: is yaml
'''

# from stuphos.lang.document.interface import load
# g = load(girl, dict(), '')
# g = Grammar.FromStructure('server', g.server)

def loadServer():
    g = loadable('WMC []\n\n' + girl).loading.structure.server
    return Grammar.FromStructure('server', g)

def test(s = g2):
    from sys import stdout
    g = loadServer()

    show(g, s, stdout)

def show(g, s, stream):
    w = stream.write

    depth = 0
    tab = '    '

    f = True
    for t in list(g.tokenize(s.lstrip())):
        if t.type == 'NEWLINE':
            f = True
            w('\n')

        elif t.type == 'INDENT':
            # w('INDENT')
            depth += 1
        elif t.type == 'DEDENT':
            # w('DEDENT')
            depth -= 1

        elif t.type == 'YAML':
            w('%s%s:\n' % (tab * depth, t.name.value))
            w(indent(t.value, tab = tab * (depth + 1)))
            w('\n') # needed?

        elif t.type != 'ENDMARKER':
            if f:
                w(tab * depth)
                f = False

            # endmarker..?
            w(str(t.value))

            if getattr(t, 'reserved', False):
                w(' ')



'The player module is a relative of interaction, but contains just player persistant stuff.'
import stuphmud.server, stuphmud.server.player.db

# __all__ = ['scan', 'format', 'RecordObject', 'page', 'loadPlayerFile', 'DBRecord']
# import new

from os.path import exists
from pprint import pprint, pformat
import imp

try:
    from pysqlite2.dbapi2 import connect as sqlite_connect
except:
    def sqlite_connect(*args):
        raise NotImplementedError

from . import playerfile
from stuphmud.server.player.db.playerfile import *

DEFAULT_PLRFL = 'etc/players'
EXTENDED_BASE = 'plrfiles'

## Implement scanning tools
def scan():
    'Generates a source of player tuples.'
    from stuphmud.server.player.db.playerfile import (read_player_entries, DEFAULT_PLRFL)
    return iter(read_player_entries(DEFAULT_PLRFL))

def format(i, ncol, fmt = '%-15.15r'):
    """
    Function that beautifies a sequence of player records i over f(or)m(a)t
    by generating lines of ncol(umns).

    """

    fmt = ' '.join([fmt] * ncol)

    for plr in i:
        if type(plr) is sequence:
            yield fmt % tuple(plr[:ncol])
        elif instanceof(plr, PlayerFile):
            yield fmt % tuple(plr.record[:ncol])
        else:
            raise TypeError # or pass

## Implementation of individualized player files:
# Pickled data that is saved to file by name in lib/plrfiles/X-X/name.pkl
# It can be installed into a player's avatar on world.enterGame.
#
# Make addressable by mud.persistance?  The registry mechanism could call
# back to the plrfiles FS defined here.  Configure getPlayerFileName

class PlayerFile(object):
    'map(PlayerFile, scan())'

    ## For looking up player-files.
    initialMap = {
        'a':'A-E', 'b':'A-E', 'c':'A-E', 'd':'A-E', 'e':'A-E',
        'f':'F-J', 'g':'F-J', 'h':'F-J', 'i':'F-J', 'j':'F-J',
        'k':'K-O', 'l':'K-O', 'm':'K-O', 'n':'K-O', 'o':'K-O',
        'p':'P-T', 'q':'P-T', 'r':'P-T', 's':'P-T', 't':'P-T',
        'u':'U-Z', 'v':'U-Z', 'w':'U-Z', 'x':'U-Z', 'y':'U-Z',

        'A':'A-E', 'B':'A-E', 'C':'A-E', 'D':'A-E', 'E':'A-E',
        'F':'F-J', 'G':'F-J', 'H':'F-J', 'I':'F-J', 'J':'F-J',
        'K':'K-O', 'L':'K-O', 'M':'K-O', 'N':'K-O', 'O':'K-O',
        'P':'P-T', 'Q':'P-T', 'R':'P-T', 'S':'P-T', 'T':'P-T',
        'U':'U-Z', 'V':'U-Z', 'W':'U-Z', 'X':'U-Z', 'Y':'U-Z',

        # 'z':'U-Z', 'Z':'U-Z',
    }

    from pickle import load as loadext, dump as dumpext
    from os.path import join

    loadext = staticmethod(loadext)
    dumpext = staticmethod(dumpext)
    join = staticmethod(join)

    def pack(self, record = None):
        import stuphmud.server.player.db.playerfile as plr
        return plr.pack(plr.pfilel_fmt, record or self.record)

    def __init__(self, record = None, player = None):
        self.record = record
        self.player = player

    def extendedFilename(self, player = None):
        ## Load a resource file into the avatar records.
        player = player or self.player

        if name is None:
            if type(player) is world.peer:
                player = player.avatar
            if type(player) is world.mobile:
                name = player.name

        assert name
        return join(EXTENDED_BASE, self.initialMap.get(name.lower()[0], 'ZZZ'), name, '.pkl')

    def applyExtendedRecord(self, record):
        'A sub-procedure of loading.'
        if record is None:
            record = self.record = self.loadExtendedRecord()

        player = self.player

        if type(record) is dict:
            # Update avatar dictionary
            for (k, v) in record.items():
                if not k.startswith('_'):
                    setattr(player, k, v)
        elif type(record) not in (list, generator):
            # Rebind pack/tobuffer methods to record-cell
            packed = str(record)
            buf = buffer(packed)

            def pack():
                return packed
            def tobuffer():
                return buf

            self.pack = pack
            self.tobuffer = tobuffer
        else:
            record = tuple(record)

        if type(record) is tuple:
            # Rebind pack/tobuffer methods to record-cell 
            packrec = self.pack
            rec2buf = self.tobuffer

            def pack():
                return packrec(record)
            def tobuffer():
                return rec2buf(record)

            self.pack = pack
            self.tobuffer = tobuffer

    def loadExtendedRecord(self, player = None):
        '''
        playerfls = map(mud.player.db.PlayerFile, mud.player.db.scan())

        p = playerfls[0]
        p.applyExtendedRecord(p.loadExtendedRecord())

        '''

        return self.loadext(self.extendedFilename(player or self.player))

    def dumpExtendedRecord(self, player = None, record = None):
        fl = open(self.extendedFilename(player or self.player), 'w')
        self.dumpext(fl, record or self.record)
        fl.flush()
        fl.close()

    def __repr__(self):
        
        d = {}

        for (n, v) in zip(mud.player.db.playerfile.pfilel_names, self.record):
            d[n] = v

        from pprint import pformat as pf
        pf = pf(d).strip()

        return '<Player %r %s>' % (self.player and self.player.name or '', pf)

# Todo: subclass PlayerFile with a positional element interfacing the position-dependent
# functions in mud.player.db.playerfile with the positional record for full database access.

def page(peer, ncol = 6):
    'Peer-friendly output for paging the results of a single scan.'
    peer.page('\r\n'.join(map(str, format(scan(), ncol))))


## Implementation of SQL save-points.
CORRUPT_SAVE_POINT_DBFL = 'etc/players-corrupt.db'

def playerSavePoint_1(player, i):
    store_tuple = unpack_player(buffer(player.store))

    e = not exists(CORRUPT_SAVE_POINT_DBFL)
    c = sqlite_connect(CORRUPT_SAVE_POINT_DBFL)
    x = c.cursor().execute

    if e:
        # print 'No file %r, creating:' % CORRUPT_SAVE_POINT_DBFL
        # print playerfile.NEW_PLAYERS_TABLE

        x(playerfile.NEW_PLAYERS_TABLE)

    # print 'Executing and commiting SQL:'
    # print playerfile.NEW_PLAYER
    # print pformat(store_tuple)[:100]

    x(playerfile.NEW_PLAYER, store_tuple)
    c.commit()

    del x
    c.close()


## SQLObject interface.

from os.path import abspath

try:
    from sqlobject import sqlhub, connectionForURI
except:
    def connectionForURI(*args):
        raise NotImplementedError

DEFAULT_SAVEPOINT_DBFL = 'etc/players-sqlobject-2.db'

def cstring(s):
    i = s.find('\x00')
    if i >= 0:
        return s[:i]
    return s

def store2dict(store):
    
    store_tuple = unpack_player(buffer(store))

    store_dict = {}
    for (n, v) in zip(playerfile.pfilel_names, store_tuple):
        store_dict[n] = v

    # Cleanse
    store_dict['skills']    = ''
    store_dict['talks']     = ''
    store_dict['affected']  = ''
    store_dict['color']     = ''

    # host should already be in clean form (take cstring then)
    store_dict['host']      = cstring(store_dict['host'])
    store_dict['name']      = cstring(store_dict['name'])
    store_dict['email']     = cstring(store_dict['email'])
    store_dict['pwd']       = cstring(store_dict['pwd'])

    store_dict['lastlogin'] = int(store_dict.get('lastlogin') or 0)

    return store_dict

def openPlayerStore(protocol = 'sqlite', path = DEFAULT_SAVEPOINT_DBFL, reload_schema = False):
    path = abspath(path)
    # e = exists(path)

    sqlhub.processConnection = \
        connectionForURI('%s:%s' % (protocol, path))

    from . import sql

    if reload_schema:
        imp.reload(sql)

    if not sql.PlayerStore.tableExists():
        sql.PlayerStore.createTable()

    return sql.PlayerStore

def playerSavePoint(player, i):
    if 1:
        save_point = store2dict(player.store)
        save_point['save_point_indication'] = str(i)

        if player.peer:
            save_point['host'] = player.peer.host

        playerSave = openPlayerStore('sqlite', DEFAULT_SAVEPOINT_DBFL)

        # Save record
        playerSave(**save_point) # .id

    world.corelog('SAVE-POINT [%s (%s)]' % (player.name, i))

def loadPlayerStore(name, store, pfilepos):
    # Search Youngest Save Points (PlayerSave table) and extract information into store.
    # Do grouping sort into record-ids with compound select?
    #
    # playerStore = openPlayerStore()
    # lastSavePoint = playerStore.selectBy(name = name, orderBy = 'lastlogon', limit = 1).reversed()
    #
    # columns = playerStore.sqlmeta.columns.keys() or playerfile.pfilel_fmt
    # for n in columns:
    #   if hasattr(store, n):
    #       setattr(store, n, getattr(lastSavePoint, n))
    #
    #
    world.corelog('LOAD-STORE [%s (position #%d; length %d)]' % (name, pfilepos, len(buffer(store))))
    # return store



# Game Core Replacement
# (C) 2021 runphase.com .  All rights reserved.
#
from stuphos.etc.tools import isYesValue

from . import Heartbeat, Game
from .db import Database

from time import time as getSystemTime
from pdb import runcall, set_trace as enter_debugger
from os import getpid, getenv
from queue import Empty
import sys

# System Database.
class Local(Database):
    sys = sys

    def __init__(self, *args, **kwd):
        # Obselete, just configure packages.
        # try: libdata = kwd['libdata']
        # except KeyError: pass
        # else:
        #     del kwd['libdata']
        #     if libdata not in self.sys.path:
        #         self.sys.path.append(libdata)

        Database.__init__(self, *args, **kwd)

    def activate(self, engine):
        engine.db = self
        self.boot()

# Application.
def freeClient(network, port):
    print('Opening Loopback Connection:', port)
    network.openConnection(('localhost', port))

def parseArgvToKwd(args):
    kwd = dict()
    for a in args:
        (name, value) = a.split('=', 1)
        kwd[name] = value

    return kwd

class Core(Heartbeat):
    class Pulse(Heartbeat.Task):
        # This could go in system.heartbeat
        from time import time as getTimeInSeconds, sleep
        getTimeInSeconds = staticmethod(getTimeInSeconds)
        sleep = staticmethod(sleep)

        OPT_MSEC = 100 # Milliseconds per timeslice.

        def getRemainingTimeslice(self, duration):
            return duration % self.OPT_MSEC # XXX what if duration > OPT_MSEC??
        def sleepForMilliseconds(self, duration):
            duration /= 1000
            # duration /= 10 # seems more accurate, but wtf?
            # print duration
            self.sleep(duration) # / 1000) # XXX :skip: don't divide by 1000!?
        def getTimeInMilliseconds(self):
            return self.getTimeInSeconds() * 1000

        def __init__(self, heartbeat):
            self.heartbeat = heartbeat

        def perform(self, engine):
            start = self.getTimeInMilliseconds()

            # import pdb; pdb.set_trace()
            # pdb.runcall(self.heartbeat.pulse, 0, 100)

            # import time, datetime
            # print datetime.datetime.fromtimestamp(time.time())

            try:
                self.heartbeat.pulse(0, 100, blocking = engine.blockingQueue)

                duration = self.getTimeInMilliseconds() - start
                remaining = self.getRemainingTimeslice(duration)
                if remaining:
                    self.sleepForMilliseconds(remaining)

            except Empty:
                # What are we catching if not an errant exception coming from
                # pulse.  This is able to clear the error which seems to not
                # be happening in the emulated machine.  It is unclear if this
                # catch will work if not for the sleep call that might raise
                # the pending Empty exception.
                pass

            except KeyboardInterrupt:
                @engine.event.call
                def done():
                    # debugOn()
                    engine.stop()
                    StuphMUD.ShutdownGame()
                    # gc.collect(); os.kill(signal.SIGTERM, os.getpid())

            except:
                from traceback import print_exc
                print_exc()

    def __init__(self, optArgs,
                 stuphlib, worldModule,
                 globalize = False,
                 initForCore = None,
                 consoleClass = None):

        (options, args) = optArgs

        if globalize:
            global instance
            import builtins as builtin
            instance = builtin.core = self

        # Configure for continuous metal.
        if options.fast_vm:
            timeout = None
        else:
            # Or blocking game queue.
            timeout = options.timeout
            if not timeout: # If not timeout is specified: burnout!
                timeout = False


        Heartbeat.__init__(self)
        self.cmdln = dict(options = options, args = args)

        # Configure against options.
        try: import readline
        except ImportError: pass

        if options.debug > 3:
            enter_debugger()

        if options.runpid:
            with open(getenv('RUNPID_FILENAME', 'runpid'), 'w+b') as o:
                o.write(str(getpid()))


        if options.local:
            # Bespoke configuration.
            options.agent_system = options.local
            options.services.insert(0, 'agentsystem=ph.interpreter.mental.library.model.localAgentSystem')


        self.headless = options.headless
        self.consoleClass = consoleClass

        if not options.headless:
            if consoleClass is not None:
                console = consoleClass.create(debug = options.debug) # interactive = options.interactive
                self += console

                if options.supreme:
                    console.avatar.level = console.avatar.LVL_SUPREME

                    # Enable 'examine' command.
                    from world.player import EnableEMHW
                    EnableEMHW() # Emergency Mode Holographic Wizard

                if options.admin_name:
                    console.avatar.name = options.admin_name
                # if options.enter_game:
                #     from stuphos.system.api import world
                #     console.avatar.room = world.room(3001)

            else:
                console = None

        self += Game(timeout = timeout)

        if options.port:
            from stuphos.system.network import MotherSocket
            network = MotherSocket(options.port)
            self += network
            # self.event.call(freeClient, self.network, options.port)

            from world import player
            player.emergeInternalPeer = network.emergeInternalPeer

            if not options.headless:
                console.attachNetwork(network)


        # Before this there is no configuration.
        if options.debug > 2:
            runcall(self.bootMudStart, options, stuphlib, worldModule)
        else:
            self.bootMudStart(options, stuphlib, worldModule)


        # Now 'configuration' builtin is available.


        # worldModule.player.EnableEMHW
        from world.player import EnableEMHW

        if isYesValue(configuration.Interpreter.emhw): # or options.emhw:
            EnableEMHW()

        if not options.headless:
            def ehwm(console):
                # Console might be None. (?)
                if console is not None:
                    console.state = 'Playing' # 'Shell'

                    if options.admin_command:
                        # todo: do this after enter game?
                        console.input = options.admin_command
                        # if options.single_command_exit:
                        #     console.input = 'exit'

                    StuphMUD.GreetPlayer(console)
                    StuphMUD.NewIncomingPeer(console)

                    try: enterGameInRoom(console, console.avatar)
                    except ValueError as e:
                        print(f'[EHWM] Enter Game: {e}')

            self.event.call(ehwm, console)

        if options.data_dir:
            from os import chdir
            print('Changing to data directory: %s' % options.data_dir)
            chdir(options.data_dir)

        if callable(initForCore):
            # Let calling context initialize, now that most of the
            # core is booted, before the main process is run.
            initForCore(self)


        if options.file:
            # XXX :skip: The file argument should be accessible with natives
            # as well as the runModuleScript/open file routines such
            # that the AgentSystem:initialization script can decide
            # when to run test console code (not very useful in a server).

            from ph.interpreter.mental import runModuleScript
            with open(options.file) as o:
                runModuleScript(o.read(), programmer = options.admin_name,
                                tracing = options.admin_trace,
                                tokenize = options.parser_tokenize)


        if options.admin_script:
            from stuphos.runtime.architecture.lookup import LookupObject
            adminScript = LookupObject(options.admin_script)
            adminScript(**parseArgvToKwd(options.admin_script_args))


        optimisticTermination = not bool(options.file) and not bool(options.admin_command)


        # The engine will be run another way.  Just return Core.
        if not options.no_run_engine:
            if options.debug > 1:
                runcall(self.run) # Todo: ShutdownGame
            elif options.asynchronous:
                from _thread import start_new_thread as nth
                nth(self.run, ()) # Todo: ShutdownGame
            else:
                try: self.run(optimisticTermination = optimisticTermination)
                except KeyboardInterrupt:
                    print()

                StuphMUD.ShutdownGame()
                # system.core.unreachable = True

    @property
    def blockingQueue(self):
        # Todo: if console is enabled, do not block!
        return self.cmdln['options'].blocking

    def resetWorldOption(self, options):
        if isYesValue(configuration.World.reset_on_boot):
            return True

        return options.reset_world

    def bootMudStart(self, options, stuphlib, worldModule):
        # Initialize MUD Package.
        import stuphos

        stuphos.bootStart(options.config_file, options.set_option) # runtime.

        # Complete MUD Boot Cycle.
        self.bootStartTime = getSystemTime()
        if not options.no_world:
            self.event.call(self.bootWorld, options, stuphlib, worldModule)

        self.event += self.worldResetStart

        if not options.no_world and self.resetWorldOption(options):
            self.event.call(self.resetWorld)

        self.event += self.worldResetComplete

        self.event += self.bootMudComplete

    def bootMudComplete(self):
        import stuphos
        elapsed = getSystemTime() - self.bootStartTime
        stuphos.bootComplete((elapsed / 1000, elapsed % 1000))

        if self.cmdln['options'].fast_vm:
            from ..metal import vm
            self.fastVM = vm()
            nth(self.fastVM)

        else:
            # Install timing driver.
            try: self += self.Pulse(stuphos.getHeartbeat())
            except AttributeError as e:
                print(f'bootMudComplete: {e}') # No need to print the traceback.

            # If components.system.core.Core construction fails
            #     to bootStart because the components.runtime.core
            #     fails, then it doesn't not fail the system core
            #     construction, it just causes the runtime.bootStart
            #     to return None (so no bridge...) 

            #     This means that later functionality may fail to
            #     load, for example, the compartmental framework
            #     heartbeat driver activated at the end of the
            #     boot protocol.

            #     Really what this means is that the boot start
            #     shouldn't fail where the boot complete continues,
            #     except that as a discrete and specific failure,
            #     a non-existence bridge should indicate no heartbeat
            #     module, and thus an alternative pulse activation.

            #         But to do this means that the core logics need
            #         to recognize and apply state.


    # def initConsole(self, console):
    #     from stuphmud.server.player import interpret
    #     interpret(console)

    def bootWorld(self, options, stuphlib, worldModule):
        if options.world_dir:
            index = 'index' if options.full_world else options.zone_index

            self += Local(stuphlib, worldModule,
                          options.world_dir, index,
                          cascade = options.cascade,
                          verbose = options.verbose)
                          # libdata = options.libdata)

    def resetWorld(self):
        return self.db.resetWorld(self, forceAll = True)

    def worldResetStart(self):
        StuphMUD.StartWorldReset()
    def worldResetComplete(self):
        # debugOn()
        StuphMUD.CompleteWorldReset()


def loadIntoRoom(peer, actor, room):
    if room is not None:
        actor.room = room
        from stuphos.system.game.namespace import lookAtRoom # import shelter
        lookAtRoom(actor, peer, room)

def enterGameInRoom(peer, actor):
    from stuphmud.server.player import getFirstValidRoom, getPlayerLoadRooms
    loadIntoRoom(peer, actor, getFirstValidRoom(getPlayerLoadRooms(actor)))
    StuphMUD.EnterGame(peer, actor)

    # from stuphmud.server.player import enterGame
    # enterGame(peer, actor)


# System Core Bridged Events.
StuphMUD.ShutdownGame = 'shutdownGame'

StuphMUD.StartWorldReset = 'resetStart'
StuphMUD.CompleteWorldReset = 'resetComplete'

StuphMUD.StartZoneReset = 'startZoneReset'
StuphMUD.CompleteZoneReset = 'completeZoneReset'

StuphMUD.MobileFromRoom = 'removeMobileFromRoom'
StuphMUD.MobileToRoom = 'putMobileInRoom'




# Player Event Framework.
class Trigger:
    def triggeredBy(self, event):
        pass
    def invokeTrigger(self, player):
        pass

class Event:
    def __init__(self, name, *args, **kwd):
        pass

def TriggerPlayerEvent(player, name, *args, **kwd):
    try: programs = player.programs
    except AttributeError: pass
    else:
        # This is lame: it should search through triggers and test them.
        try: trigger = programs[name]
        except (TypeError, KeyError): pass
        else:
            if isinstance(trigger, Trigger):
                # if trigger.triggeredBy(event)
                if player.implementor:
                    trigger.invokeTrigger(player, *args, **kwd)



# Service/Daemon Management.

# todo: provide an application provider interface
from sys import platform

# if platform == 'cygwin':
from os import kill, spawnv, P_NOWAIT
from signal import SIGKILL as KILL_SIGNAL, SIGQUIT as QUIT_SIGNAL
from errno import ESRCH

from optparse import OptionParser, Option

from stuphos.runtime.facilities import Facility
from stuphos.runtime import Component
from .config import loadConfig
from ..etc import parseOptionsOverSystem, fromJsonFile, OpenFile, FileNotFoundException
from stuphos import getSection

import signal
_sigcodes = dict((getattr(signal, sig), sig) for sig in dir(signal) \
                 if sig.startswith('SIG'))

def parseCmdln(argv, *args):
    parser = OptionParser()

    for param in args:
        if type(param) in (list, tuple):
            parser.add_option(*param)
        elif isinstance(param, str):
            parser.add_option(param)
        elif isinstance(param, Option):
            parser.add_option(param)

    if argv in (list, tuple):
        argv = ' '.join(map(str, argv))

    return parseOptionsOverSystem(parser, argv)

def getStatusFile(folder, name):
    return joinpath(folder, name) + '-status.json'

def loadStatusFile(folder, name):
    def _():
        try: return fromJsonFile(OpenFile(getStatusFile(folder, name)))
        except FileNotFoundException: return dict()

    return SubDaemonManager.SubDaemon.Status(**_())

def saveStatusFile(folder, name, data):
    with open(getStatusFile(folder, name), 'w') as outfile:
        toJsonFile(data, outfile)

class SubDaemonManager(Facility):
    # This should probably use the subprocess module to handle child exits.
    NAME = 'SubDaemon::Manager'

    class Manager(Facility.Manager):
        VERB_NAME = 'subdaemon-*manage'
        MANAGE_LEVEL = Facility.Manager.IMPLEMENTOR

        USAGE = 'Usage: subdaemon-manage { load | unload | show [--long] | start | quit [--abort] | reload }'

        def doCommand(self, peer, cmd, args):
            if peer.avatar and peer.avatar.level >= self.MINIMUM_LEVEL:
                largs = args and args.strip().lower() or ''
                args = args and args.split() or ()

                try:
                    # Re-mapping facility-management sub-commands:
                    if not largs:
                        print(self.facility.get() or 'Not installed.', file=peer)
                    elif largs == 'load':
                        print(self.facility.get(create = True), file=peer)
                    elif largs == 'unload':
                        print(self.facility.destroy() and \
                              'Destroyed.' or 'Unknown facility.', file=peer)

                    # Use these ones for the subdaemon control.
                    else:
                        this = self.facility.get()
                        if this is None:
                            print('Facility is not loaded.', file=peer)
                        elif not self.doSubCommand(this, peer, args):
                            print(self.USAGE, file=peer)

                except RuntimeError as e:
                    print(e, file=peer)
                except SystemExit:
                    pass

                return True

        def doSubCommand(self, mgr, peer, args):
            # Q: Hasn't this been factored into the Manager base class?
            name = 'do_%s' % (args[0],)
            subcmd = getattr(self, name, None)

            if callable(subcmd):
                subcmd(mgr, peer, args[1:])
                return True

        def getDaemon(self, mgr, args):
            try: name = args[0]
            except KeyError:
                raise RuntimeError('Must specify <subdaemon-name>!')

            try: return mgr.getDaemon(name)
            except NameError:
                raise RuntimeError('Unknown daemon: %r' % name)

        def do_start(self, mgr, peer, args):
            daemon = self.getDaemon(mgr, args)
            if not daemon.isRunning():
                daemon.startProcess(args, manager = self)
                print(daemon, file=peer)
            else:
                print('Already running: %r' % daemon, file=peer)

        def do_quit(self, mgr, peer, args):
            (options, args) = parseCmdln(args, Option('--abort', action = 'store_true'))
            daemon = self.getDaemon(mgr, args)
            if daemon.isRunning():
                if options.abort:
                    daemon.abortProcess(manager = self)
                else:
                    daemon.quitProcess(manager = self)

                print(daemon, file=peer)
            else:
                print('Not running: %r' % daemon, file=peer)

        def do_show(self, mgr, peer, args):
            (options, args) = parseCmdln(args, Option('--long', action = 'store_true'))

            if args:
                daemons = []
                for name in args:
                    try: selected.append((name, mgr.getDaemon(name)))
                    except NameError:
                        raise RuntimeError('Unknown daemon: %r' % name)
            else:
                daemons = iter(mgr.daemons.items())

            if options.long:
                daemons = ('%s\r\n%s\r\n%s\r\n' % \
                           (name, '-' * len(name),
                            daemon.getLongDescription()) \
                           for (name, daemon) in daemons)
            else:
                daemons = ('%s: %r' % nd for nd in daemons)

            peer.page_string('\r\n'.join(daemons) + '\r\n')

        def do_reload(self, mgr, peer, args):
            mgr.reloadAllConfigs()
            print(mgr, file=peer)

        # detach, pause, resume

    class SubDaemon:
        # Todo: provide a subprocess architecture for piping i/o.
        RAW = ['arguments']

        class Status:
            def __init__(self, pid = -1, state = None, exepath = None):
                self.pid = pid
                self.state = state
                self.exepath = exepath

            def getSavedData(self):
                return dict(pid = self.pid,
                            state = self.state,
                            exepath = self.exepath)

            def synchronize(self, daemon):
                # Todo: Introspect the process at pid and validate that
                # it matches the exepath/start-time to confirm state.

                # daemon.pid = self.pid
                # daemon.state = self.state

                # Todo: if the daemon is found to not be started, attempt
                # to start it now?  XXX How to do this without manager?
                #
                #     Have this function return an outcome code to signal
                #     this post-condition to the loading manager.

                pass

        @classmethod
        def parseConfig(self, section):
            for opt in section.options():
                if opt in self.RAW:
                    get = section.config.config.get
                    yield (opt, get(section.section, opt, raw = True))
                else:
                    yield (opt, section.get(opt))

        def __init__(self, name, config):
            self.name = name
            self.config = config
            self.pid = -1
            self.state = 'unstarted'

        def getPid(self):
            return self.pid
        def getState(self):
            return self.state
        def isRunning(self):
            return self.getState() in ['running', 'started']
        def isPersistant(self):
            return self.config.get('persistant', False)

        try:
            DEFAULT_QUIT_SIGNAL = QUIT_SIGNAL
            DEFAULT_ABORT_SIGNAL = KILL_SIGNAL
        except NameError:
            # no platform....
            pass

        @classmethod
        def parseSignal(self, signame):
            if isinstance(signame, int):
                assert signame in _sigcodes
                return signame

            if isinstance(signame, str):
                signame = signame.upper()
                if signame.startswith('SIG'):
                    return getattr(self.signal, signame)

            raise ValueError(signame)

        def getExePath(self):
            return self.config['exepath']
        def getQuitSignal(self):
            return self.parseSignal(self.config.get('quit-signal', self.DEFAULT_QUIT_SIGNAL))
        def getAbortSignal(self):
            return self.parseSignal(self.config.get('abort-signal', self.DEFAULT_ABORT_SIGNAL))
        def getCommandLine(self, args):
            # todo: be able to inject args into command-line:
            #    detect command-line interpolatable values
            #    generate option parser from these values
            #    parser args with these options
            #    plug values into evaluation

            cmdln = [self.getExePath()]
            template = self.config.get('arguments', '')
            argv = template % self.config
            cmdln.extend(argv.split())

            return cmdln

        def saveStatus(self, manager):
            if manager is not None:
                manager.saveStatusFile(self.name,
                                       self.Status(pid = self.pid,
                                                   status = self.state,
                                                   exepath = exepath))

        def synchronize(self, loadStatus):
            status = self.loadStatus(self.name)
            status.synchronize(self)

        # Process life-cycle.
        def startProcess(self, args, manager = None):
            exepath = self.getExePath()
            args = self.getCommandLine(args)

            self.cmdln_instance = args
            self.pid = spawnv(P_NOWAIT, exepath, args)
            self.state = 'started'

            self.saveStatus(manager)

        def killProcess(self, abort = False, manager = None):
            killSignal = self.getAbortSignal() if abort else self.getQuitSignal()
            pid = self.getPid()

            if pid >= 0:
                try: kill(pid, killSignal)
                except OSError as e:
                    if e[0] == ESRCH:
                        from stuphos.system.api import mudlog
                        mudlog('No such process: %d' % pid)

            self.state = 'killed'
            self.saveStatus(manager)

        def quitProcess(self, manager = None):
            return self.killProcess(manager = manager)
        def abortProcess(self, manager = None):
            return self.killProcess(abort = True, manager = manager)

        def restartProcess(self, args = (), abort_ok = False, manager = None):
            if abort_ok:
                self.abortProcess(manager = manager)
            else:
                self.quitProcess(manager = manager)

            self.startProcess(args, manager = manager)

        def getLongDescription(self):
            return '\n'.join(['Pid: %d' % self.getPid(),
                              'State: %s' % self.getState()] + \
                             ['%s: %s' % nv for nv in self.config.items()])
        def __str__(self):
            return 'exepath=%s pid=%d state=%s' % \
                   (self.getExePath(), self.getPid(), self.getState())
        __repr__ = __str__

    @classmethod
    def getAllConfigs(self):
        section = getSection('SubDaemonManager')
        if section:
            for opt in section.options():
                if opt == 'path' or opt.startswith('path.'):
                    cfg = loadConfig(section.get(opt))
                    for name in cfg.sections():
                        section = cfg.getSection(name)
                        subdaemoncfg = self.SubDaemon.parseConfig(section)
                        subdaemoncfg = dict(subdaemoncfg)
                        yield (name, self.SubDaemon(name, subdaemoncfg))

    @classmethod
    def getStatusFileDir(self):
        section = getSection('SubDaemonManager')
        if section:
            return section.get('statusfile-dir')

    @classmethod
    def create(self):
        statusfiledir = self.getStatusFileDir()
        assert statusfiledir # isdir(statusfiledir)

        return self(dict(self.getAllConfigs()),
                    statusfiledir)

    def __init__(self, daemons, statusfiledir):
        self.statusfiledir = statusfiledir
        self.installDaemons(daemons)
        Component(self.NAME, Target = self)

    def __registry_delete__(self):
        self.abortAllProcesses()

    def installDaemons(self, daemons):
        self.daemons = daemons

        # Scan the pidfiles for persistant daemons to load running state.
        for daemon in self.daemons.values():
            if daemon.isPersistant():
                daemon.synchronize(self.loadStatusFile)

    def loadStatusFile(self, name):
        return loadStatusFile(self.statusfiledir, name)
    def saveStatusFile(self, name, status):
        return saveStatusFile(self.statusfiledir, name,
                              status.getSavedData())

    def reloadAllConfigs(self):
        # Hard
        self.abortAllProcesses()
        self.installDaemons(dict(self.getAllConfigs()))

    def abortAllProcesses(self):
        for daemon in self.daemons.values():
            if not daemon.isPersistant():
                daemon.abortProcess()

    def getDaemon(self, name):
        try: return self.daemons[name]
        except KeyError:
            raise NameError(name)

    # This should be necessary -- the registry should delete.
    def onShutdownGame(self, ctlr, xxx_todo_changeme):
        (circle_shutdown, circle_reboot,
                                    shutdown_mode, shutdown_timer, shutdown_by) = xxx_todo_changeme
        self.abortAllProcesses()

    def __str__(self):
        return '%s (%d daemons)' % (self.__class__.__name__,
                                    len(self.daemons))
    __repr__ = __str__

# Install Management.
SubDaemonManager.manage()

# API -- Q: Is this necessary if it's all tucked into SubDaemonManager object??
try: from stuphos.runtime.registry import registerObject
except ImportError: pass
else:
    #@runtime.api(runtime.Node(SubDaemonManager.NAME).API)
    class API:
        def getManager(self, start_ok = False):
            return SubDaemonManager.get(create = bool(start_ok))
        def startManager(self):
            return self.getManager(start_ok = True)

        def reloadManager(self):
            return self.getManager().reloadAllConfigs()

        def getDaemon(self, name):
            return self.getManager().getDaemon(name)

        def startDaemon(self, name):
            daemon = self.getDaemon(name)
            if not daemon.isRunning():
                daemon.startProcess(args, manager = self.getManager())

        def quitDaemon(self, name, abort_ok = False):
            daemon = self.getDaemon(name)
            if daemon.isRunning():
                if abort_ok:
                    daemon.abortProcess(manager = self.getManager())
                else:
                    daemon.quitProcess(manager = self.getManager())

        def restartDaemon(self, name, args = (), abort_ok = False):
            daemon = self.getDaemon(name)
            if daemon.isRunning():
                daemon.restartProcess(args = (), abort_ok = abort_ok,
                                      manager = self.getManager())

    # Singleton.
    API = API()

    SUBDAEMON_API_OBJECT_NAME = SubDaemonManager.NAME + '::API'
    registerObject(SUBDAEMON_API_OBJECT_NAME, API)



# Error and Informational Logging Routines.
# These are loaded into runtime.core module.
from . import getSystemException, getModuleFileBasename
from traceback import extract_tb

__all__ = ['log', 'logWarning', 'logError', 'logException', 'logWizards']

def log(*args, **kwd):
    global log
    from stuphos.system.api import syslog # Ultimately: system.game.compartment
    log = syslog # Rewrite.

    # runtime.call.System.Bridge.logSystem(*args, **kwd)
    # logJournal(*args, **kwd)

    return syslog(*args, **kwd)

def logJournal(*args, **kwd):
    log = runtime[runtime.System.Journal]
    if log is not None:
        from stuphos.etc.tools.timing import date as now
        import json

        # @runtime.on.System.Bridge.logSystem
        # :system-journal:
        log += ('system', 'log', json.dumps((args, kwd)), now())

def logWarning(message):
    log('Warning: %s' % message)
def logError(message):
    log('Error: %s' % message)

DEFAULT_LOGINDENT = 27
PADINDENT = ' ' * 2

def tracebackString(tb = None, indent = None):
    if indent is None:
        indent = DEFAULT_LOGINDENT

    def _():
        logindent = indent * ' '
        logindent += PADINDENT
        logindent2 = logindent + PADINDENT

        for (file, lineno, name, source) in extract_tb(tb):
            file = getModuleFileBasename(file)
            yield '%s[%s:%d] %s' % (logindent, file, lineno, name)
            if source:
                yield '%s%s' % (logindent2, source.strip())

    return '\n'.join(_())

def exceptionHeader(etype = None, value = None, tb = None):
    return '%s: %s' % (etype.__name__, value)

def logException(etype = None, value = None, tb = None,
                 traceback = False, header = None,
                 indent = 'auto'):

    # Construct a traceback suitable for logging to syslog file.
    # Todo: serialize the exception data to another file.
    if etype is None and value is None and tb is None:
        (etype, value, tb) = getSystemException()

    if header is None:
        header = exceptionHeader(etype, value, tb)

    if traceback:
        if indent == 'auto':
            from stuphos import getConfig

            indent = getConfig('logindent')
            if indent is not None:
                indent = int(indent)


        tb = tracebackString(tb, indent = indent)

        if header:
            log('%s\n%s' % (header, tb))
        else:
            log(tb)
    else:
        log(header)

DEFAULT_WIZLOG_LEVEL = 115
DEFAULT_WIZLOG_TYPE = 'Complete'

def logWizards(message, level = None, type = DEFAULT_WIZLOG_TYPE, tofile = False):
    if level is None:
        level = DEFAULT_WIZLOG_LEVEL

    elif not isinstance(level, int):
        from world import mobile
        if isinstance(level, mobile):
            level = mobile.level
        else:
            raise TypeError(type(level).__name__)

    from stuphos.system.api import mudlog, syslog
    mudlog(message, level, type)

    if tofile:
        syslog(message)

wizlog = logWizards


class logger(object):
    '''
    log.memory.new_environment(id(memory), self.__class__.__name__)

    '''

    @classmethod
    def _nullreceiver(*parameters):
        pass

    def __init__(self, receiver):
        self._receiver = receiver
    def __call__(self, *parameters):
        return self._receiver(*parameters)

    def __getattr__(self, parameter):
        try: return object.__getattribute__(self, parameter)
        except AttributeError:
            return self._parameter(self, parameter)

    class _parameter(list):
        def __init__(self, log, *parameters):
            self._log = log
            list.__init__(self, *parameters)

        def __call__(self, *parameters):
            return self._log(*self)

        def __getattr__(self, name):
            try: return object.__getattribute__(self, name)
            except AttributeError:
                return self.__class__(self._log, *(self + (name,)))



# Extensions -- Executable.
from stuphos.etc.tools.strings import indent, nling

from base64 import b64encode
from json import dumps as toJsonString, loads as loadJsonString

PACKAGE_META_NAME = '.package.meta'

def b64_encode(s):
    # treat as bytes (for media content)
    # s = s.encode()

    return b64encode(s).decode()
# b64_encode = b64encode 

def quoteStringIf(s):
    return repr(s) if ':' in s else s

def renderOwnedContent(node, srcContent, srcOwner, destContent, destOwner, fixupContentHeading):
    progr = node.get(srcOwner)
    if progr: # Note: filter all emptinesses
        yield '%s:' % quoteStringIf(node['name'])

        yield '    %s: %s' % (destOwner, progr)
        yield '    %s::' % destContent

        yield indent(indent(node[srcContent])) # todo: fixupContentHeading?

    else:
        yield '%s::' % quoteStringIf(node['name'])
        yield indent(fixupContentHeading(node[srcContent]))

def renderPackageString(node, fixupHeading = False):
    def fixupContentHeading(content):
        return content.lstrip() if fixupHeading else content

    @nling
    def convert(node):
        type = node['type']

        if type == 'directory':
            i = node['children']

            if i:
                yield '%s:' % quoteStringIf(node['name'])

                s = [] # structures collection
                m = [] # media collection

                for c in i:
                    if c['type'] == 'structure':
                        s.append(c)
                        continue

                    elif c['type'] == 'media':
                        m.append(c)
                        continue

                    yield indent(convert(c))

                if s:
                    # yield indent('interfaces$:')
                    yield indent('interfaces:')

                    for c in s:
                        yield indent(convert(c), level = 2)

                if m:
                    # yield indent('media$:')
                    yield indent('media:')

                    for c in m:
                        yield indent(convert(c), level = 2)

            else:
                yield '%s: []' % quoteStringIf(node['name'])

        elif type == 'module':
            yield from renderOwnedContent(node,
                                          'program', 'programmer',
                                          'program', 'programmer',
                                          fixupContentHeading)

        elif type == 'structure':
            yield from renderOwnedContent(node,
                                          'document', 'programmer',
                                          'content', 'owner',
                                          fixupContentHeading)

        elif type == 'media':
            content_type = node['content_type']
            if content_type:
                yield '%s:' % quoteStringIf(node['name'])

                yield '    content_type: %s' % content_type
                yield '    content::'

                content = node['content']
                if isinstance(content, str):
                    content = content.encode()

                yield indent(indent(b64_encode(content)))

            else:
                yield '%s::' % quoteStringIf(node['name'])
                yield indent(b64_encode(node['content']))

    def rewrite(node):
        # Transform tree root node name.
        if node['name'] is None:
            node['name'] = ''

        # debugOn()
        return node

    return convert(rewrite(node.exported))


class filesystem:
    # Represent a folder tree as structural input to the upload routine.

    packageMetaName = PACKAGE_META_NAME

    def __init__(self, path, **kwd):
        self.path = path
        self.config = kwd

        try: self.packageMetaName = kwd.pop('packageMetaName')
        except KeyError: pass

    def contents(self, path):
        value = path.open('r+b').read()
        try: return value.decode()
        except UnicodeDecodeError as e:
            print(f'[pkging$filesystem$items] {path}: {e}')
            return value

        return path.read() # .platformMapped

    def __len__(self):
        return len(self.path.listing)

    def __getitem__(self, name):
        if name == self.packageMetaName:
            return self.path(self.packageMetaName).read()

        if name in ['structures', 'interfaces']:
            try: return dict((p.basename, self.contents(p)) for p in
                             self.path(name).listing if not p.isdir)

            except FileNotFoundError:
                return dict()

        raise KeyError(name)

    def ignore(self, path):
        return path.basename in self.config.get('ignore', [])

    def iteritems(self):
        try: i = self.path.listing
        except FileNotFoundError: pass
        else:
            for p in i:
                if self.ignore(p):
                    continue

                if p.isdir:
                    yield (p.basename, self.__class__(p))
                else:
                    try: yield (p.basename, self.contents(p))
                    except UnicodeDecodeError as e:
                        print(f'[pkging$filesystem$items] {p}: {e}')

    def items(self):
        return list(self.iteritems())


def packageBuild(path, **config):
    @nling
    def w(o):
        if isinstance(o, (dict, filesystem)):
            for (name, value) in o.items():
                if isinstance(value, (list, tuple, dict, filesystem)):
                    yield f'{name}:'
                    yield indent(w(value))

                elif isinstance(value, bytes):
                    pass
                elif isinstance(value, str):
                    yield f'{name}::'
                    yield indent(value)

                else:
                    yield f'{name}: {repr(value)}'

        elif isinstance(o, (list, tuple)):
            if o:
                for value in o:
                    o = w(o)
                    o = o.split('\n')
                    if len(o) > 1:
                        yield '- ' + o[0]
                        yield indent('\n'.join[1:], tab = '  ')
                    else:
                        yield '- ' + o[0]
            else:
                yield '[]'

        else:
            yield repr(o)

    return w(filesystem(path, **config))


def extractOwnedContent(s, content, owner):
    try: content = s[content]
    except KeyError: pass
    else:
        # If program is set, programmer must be set or it's added as a folder.
        try: owner = s[owner]
        except KeyError: pass
        else:
            if len(s) == 2:
                return (content, owner)


ALT_TYPES = ('structures', 'interfaces', 'media')

# Todo: ignore_package_meta is used as the package meta filename.
def uploadStructure(core, path, structure, set_programmers = False,
                    ignore = None, folders = None, attributes = None,
                    ignore_package_meta = False, srcpath = None):
    structs = []
    media = []

    if srcpath is None:
        srcpath = path

    reserved = ALT_TYPES + ((PACKAGE_META_NAME,) if ignore_package_meta is True
                            else (() if not ignore_package_meta else (ignore_package_meta,)))

    if attributes == 'load':
        name = ignore_package_meta if isinstance(ignore_package_meta, str) else PACKAGE_META_NAME

        try: attributes = structure[name]
        except KeyError as e:
            raise e
        else:
            if not isinstance(attributes, str):
                raise ValueError(f'Expected simple text for {name}, got {type(attributes)}')

            attributes = loadJsonString(attributes)

    elif not isinstance(attributes, dict):
        attributes = dict()

    def getOwner(dest, name):
        if srcpath:
            dest = dest[len(srcpath)+1:]

        try: return attributes[f'{dest}/{name}']['owner']
        except KeyError:
            pass

    try: structs.extend(list(structure['structures'].items()))
    except KeyError: pass

    try: structs.extend(list(structure['interfaces'].items()))
    except KeyError: pass

    try: media.extend(list(structure['media'].items()))
    except KeyError: pass

    services = [(name, s) for (name, s) in structure.items()
                if name not in reserved]

    def ensure(path):
        path = path.split('/')
        u = core.root.lookup

        for i in range(1, len(path)+1):
            folder = path[:i]

            try: u(*folder)
            except KeyError:
                core.addFolder('/'.join(folder[:i-1]), folder[-1])

    def isFolder(path, name):
        if folders:
            return f'{str.__str__(path)}/{str.__str__(name)}' in folders

    def install(path, struct, add, isMedia = False, isStructure = False):
        # head = (path + '/') if path else ''

        if path:
            head = path + '/'
            ensure(path)
        else:
            head = ''

        for (name, s) in struct:
            name = str(name) # Because it gets __sqlrepr__ attr which might be complex.

            if ignore and ignore(name): # todo: this should probably turn into ignore(path, name)
                continue

            if isMedia:
                if isinstance(s, str):
                    s = b64_decode(s)
                    add(path, name, s)

                elif isinstance(s, (dict, filesystem)):
                    try: content = s['content']
                    except KeyError: pass
                    else:
                        try: content_type = s['type']
                        except KeyError: content_type = None

                        content = b64_decode(content)
                        add(path, name, content, content_type)

            elif isinstance(s, str):
                ensure(path)

                # print(f'{add.__name__}({path}/{name})')

                # if name == 'system:initialize':
                #     debugOn()

                add(path, name, s, programmer = getOwner(path, name) \
                        if set_programmers else None)

            elif isinstance(s, (dict, filesystem)):
                if not isFolder(path, name):
                    ownedContent = extractOwnedContent(s, 'content', 'owner') if isStructure \
                                   else extractOwnedContent(s, 'program', 'programmer')

                    if ownedContent is not None:
                        # debugOn()
                        (content, progr) = ownedContent

                        ensure(path)

                        add(path, name, content, programmer = progr if set_programmers else None)

                        # core.addModule(path, name, content,
                        #                programmer = progr if set_programmers else None)


                        # print(f'adding module: {path} {name} [{progr}]')

                        # XXX use add? this will allow structures to set programmer/owner.
                        # core.addModule(path, name, content)
                        # if set_programmers:
                        #     core.setActivityProgrammer(path.split('/') if path else [],
                        #                                name, progr)

                        continue

                # debugOn()

                # print(f'adding folder: {path}/{name}')
                # assert not '/' in name
                if path:
                    ensure(path + '/' + name)
                else:
                    ensure(name)

                uploadStructure(core, head + name, s,
                                set_programmers = set_programmers,
                                ignore = ignore, folders = folders,
                                attributes = attributes,
                                ignore_package_meta = ignore_package_meta,
                                srcpath = srcpath) # recurse


    # st = dict(structs)
    # if 'library' in st:
    #     debugOn()

    # print(f'install {path} {", ".join}')

    install(path, structs, core.addStructure, isStructure = True)
    install(path, services, core.addModule)
    install(path, media, core.addMedia, isMedia = True)


class fsPackageCore:
    # AgentSystem replacement for extracting packages to filesystem.
    class Node:
        pass

    class Folder(Node, dict):
        def lookup(self, path, *args):
            return self[path]

    def __init__(self, path):
        self.path = io.path(path)
        self.root = self.Folder()
        self.init()

    def addFolder(self, *args, **kwd):
        print('addfolder ' + str(args))
    def addModule(self, *args, **kwd):
        print('addmodule ' + str(args))
    def addStructure(self, *args, **kwd):
        print('addstructure ' + str(args))
    def addMedia(self, *args, **kwd):
        print('addmedia ' + str(args))

    def init(self):
        pass
    def finished(self):
        pass

class fsUnpackCore(fsPackageCore):
    packageMetaName = PACKAGE_META_NAME

    def init(self):
        self.attributes = dict()

    def finished(self):
        # Overwrite.
        self.path(self.packageMetaName).write \
            (toJsonString(self.attributes))

    def setOwnerAttribute(self, path, name, owner):
        self.attributes.setdefault(f'{path}/{name}', dict()) \
            ['owner'] = owner


    def getLocalPath(self, path, name, ensure = False):
        nparts = len(self.path.parts)
        local = self.path
        for p in path.split('/') + [name]:
            local = local(p)
            if len(local.parts) < nparts:
                raise ValueError(f'{path}/{name}')

        if ensure:
            local.folder.ensure()

        return local

    def addFolder(self, path, name):
        # print('folder', self.getLocalPath(path, name, ensure = False))
        self.getLocalPath(path, name, ensure = False).ensure()

    def addModule(self, path, name, content, programmer = None):
        # print('module', self.getLocalPath(path, name, ensure = True))
        if programmer is not None:
            self.setOwnerAttribute(path, name, programmer)

        self.getLocalPath(path, name, ensure = True).write(content)

    def addStructure(self, path, name, content, content_type = None, programmer = None):
        # print('structure', self.getLocalPath(f'{path}/interfaces', name, ensure = True))
        if programmer is not None:
            self.setOwnerAttribute(path, name, programmer)

        self.getLocalPath(f'{path}/interfaces', name, ensure = True) \
            .write(content)

    def addMedia(self, path, name, content, content_type = None):
        # print('media', self.getLocalPath(f'{path}/media', name, ensure = True))
        self.getLocalPath(f'{path}/media', name, ensure = True) \
            .write(content)


def packageUnpackTo(structure, dest_dir, mount_point = None, fsUnpackClass = fsPackageCore, **kwd):
    'Output package to filesystem (or something that takes an argument).'

    if isinstance(structure, str):
        from stuphos.language.document.interface import document
        structure = document(structure)

    core = fsUnpackClass(dest_dir, **kwd)
    # debugOn()
    uploadStructure(core, mount_point or '', structure, set_programmers = True)
    core.finished()

def packageStreamUnpackTo(input, output, mount_point = None):
    '''
    --admin-script=ph.interpreter.mental.library.extensions.packageStreamUnpackTo \
    -x bin.package -x bin

    '''

    if input == '<stdin>':
        from sys import stdin as input
    else:
        input = open(input).read()

    return packageUnpackTo(input, output, mount_point = mount_point)


class normalContinue(Exception):
    def __init__(self, options, args):
        self.options = options
        self.args = args

        raise self


def unpackMain(options, args):
    if len(args) != 2:
        print('Usage: unpack <input package> <output path>')
        return

    (input, output) = args

    from stuphos.runtime.registry import getRegistry
    getRegistry(create = True) # stuphos.structure

    packageStreamUnpackTo(input, output, options.mount_point)


class project:
    def __init__(self, path, config, options):
        if config == '.':
            config = '.wmc'

        self.path = path = io.path(path)
        self.configpath = path(config)
        self.options = options

    @property
    def config(self):
        return self.configuration(self.path.config.cfgOf.value)

    class configuration(dict):
        def __init__(self, fileInputValue):
            pass

    def package(self, args):
        self.options.input_file = str(self.path)
        normalContinue(self.options, args)

    def upload(self, args):
        cfg = self.cfg


    @classmethod
    def Operate(self, options, args):
        assert args
        p = project(io.here, options.project, options)

        cmd = args[0]
        rest = args[1:]

        if cmd == 'upload':
            return p.upload(rest)
        elif cmd == 'package':
            return p.package(rest)


def restricted(self, *args):
    r = self

    for a in args:
        if isinstance(a, str) and not isinstance(a, io.path):
            a = io.path(a)

        for p in a.parts:
            if p == '..':
                if len(r) > len(self):
                    r = r.folder

            elif p != '.':
                r = r(p)

    return r

ContentIsFolder = object()

from collections import namedtuple

updateFSOpTuple = namedtuple('UpdateFSOp', 'base path content type')
updateFSTypeTuple = namedtuple('UpdateFSType', 'name commit')

def updateFSTree(base, files):
    base = io.path(base)
    for (name, content) in files:
        path = restricted(base, name)

        if content is ContentIsFolder:
            def op(base, path, content):
                def buildFolder():
                    path.ensure()

                return ('folder', buildFolder)

        else:
            def op(base, path, content):
                def writeFileContent():
                    path.ensure(folder_only = True)
                    path.write(content)

                return ('file', writeFileContent)

        yield updateFSOpTuple(base, path, content, \
               updateFSTypeTuple(*op(base, path, content)))


# @apply
# class defaultIgnore:
#     def __contains__(self, object):
#         return object in ['.git']

defaultIgnore = ['.git']


USAGE = \
'''
export PYTHONPATH=<path-to-common>:<path-to-stuphos>
python -m stuphos.management.packaging -o jhcore.package path/to/LambdaMOO/core.db

'''.rstrip()

def main(argv = None):
    # debugOn()
    from optparse import OptionParser
    parser = OptionParser(usage = USAGE)
    parser.add_option('-o', '--output-file', '--output')
    parser.add_option('-i', '--input-file', '--input')
    parser.add_option('-u', '--unpack', action = 'store_true')
    parser.add_option('--mount-point')
    parser.add_option('-p', '--project')
    parser.add_option('-v', '--verbose', action = 'store_true')
    (options, args) = parser.parse_args(argv)

    try: import op
    except Exception as e:
        print(f'{e.__class__.__name__}: {e}')

    try:
        if options.project:
            return project.Operate(options, args)

        if options.unpack:
            return unpackMain(options, args)

    except normalContinue as e:
        options = e.options
        args = e.args

    if options.input_file:
        assert not args
        input = options.input_file
    else:
        assert len(args) == 1
        input = args[0]

    if options.output_file:
        output = open(options.output_file, 'w')
    else:
        from sys import stdout as output

    import op

    output.write(str(packageBuild \
        (io.path(input),
         ignore = defaultIgnore)))

if __name__ == '__main__':
    main()



# Compartmental Command-line Bootstrap Interface.
from optparse import OptionParser
import sys, os

try: import pdb
except ImportError: pass

def parseCmdln(argv = None):
    parser = OptionParser()

    parser.add_option('-C', '--config-file', '--config', '--game-config')
    parser.add_option('-g', '--debug', action = 'count', default = 0)
    parser.add_option('-p', '--port', type = int)
    parser.add_option('-v', '--verbose', action = 'count', default = 0)
    parser.add_option('-d', '--headless', '--no-console', action = 'store_true')

    parser.add_option('-w', '--world-dir')
    parser.add_option('-z', '--zone-index', '--index')
    parser.add_option('--reset-world', action = 'store_true')
    parser.add_option('-Z', '--full-world', action = 'store_true')
    parser.add_option('-n', '--no-world', action = 'store_true')
    parser.add_option('-W', '--cascade', '--load-world', action = 'store_true')

    parser.add_option('-i', '--interactive', action = 'store_true')
    parser.add_option('-a', '--asynchronous', '--async', action = 'store_true')
    parser.add_option('-m', '--mud-package', '--mud')
    parser.add_option('-e', '--service', action = 'append', default = [], dest = 'services')
    parser.add_option('-S', '--set-option', action = 'append', default = [])

    parser.add_option('-s', '--supreme', action = 'store_true')
    parser.add_option('--enter-game', action = 'store_true')

    parser.add_option('--admin-name')

    # These are different than admin-name:
    parser.add_option('--admin-script')
    parser.add_option('-x', '--admin-script-args', default = [], action = 'append')

    parser.add_option('-f', '--file') # Like -c, but a whole file as module script.

    # And this is for injecting initial commands into a console-based session.
    parser.add_option('-c', '--admin-command')
    parser.add_option('--admin-trace', action = 'store_true')

    parser.add_option('--parser-tokenize', action = 'store_true')
    parser.add_option('--elemental-parser-debug', '--parser-debug')

    parser.add_option('-L', '--data-dir', '--lib-dir')
    # parser.add_option('--libdata')

    parser.add_option('--fast-vm', action = 'store_true')
    parser.add_option('--blocking', type = int, default = 0)
    parser.add_option('--timeout', type = int, default = 0)
    parser.add_option('--runpid', action = 'store_true')

    parser.add_option('--no-run-engine', action = 'store_true')
    parser.add_option('--no-init', action = 'store_true')

    parser.add_option('--tool-package', action = 'append', default = [])
    parser.add_option('--load-fixture', action = 'append', default = [], dest = 'fixture_load_set')
    parser.add_option('--agent-system')
    parser.add_option('--local', '--local-filesystem')

    return parser.parse_args(argv)


def getCmdln(argv = None):
    return sys.argv if argv is None else argv


    # can't specify a --mud-package to the game.cli because the game module
    # now fully imports the compartment system necessarily, which depends on
    # the stuphos.runtime package.

    #     The implications of this are that the compartment server cannot
    #     boot independently or independently load a particular application
    #     base package, but rather is must essentially statically link with
    #     the runtime.  The only real reason for this is because: the cli
    #     application bootstrap code originally loaded runtime components
    #     AT runtime (dynamically), but because it's doing so using the game
    #     module, and everything in the synthetic game module is now statically
    #     imported so that there's no relative package imports using it's name,
    #     it's because of the static nature of the synthetic game module that
    #     we can back everything with compartment concretely but update very
    #     little about the way we refer to game, as a container concept.


class ConsoleBusiness:
    # Heartbeat Task Implementation.
    def activate(self, engine):
        self.engine = engine
        engine.console = self
    def deactivate(self, engine):
        try:
            if self.engine is engine:
                del self.engine
        except AttributeError:
            pass

        try:
            if engine.console is self:
                del engine.console
        except AttributeError:
            pass

    def perform(self, engine):
        self.engine.event += self.handleConsoleInput

    # Game I/O Events.
    PROMPT = ' --+====> ' # ' +> '

    def readConsoleInput(self):
        # Todo: Do not perform.  Instead, enable an XMLRPC method for entering input.
        return input(self.prompt or self.PROMPT).rstrip()

    def handleConsoleInput(self):
        'Interpret player input as command for avatar.'

        try:
            self.forceInput(self.readConsoleInput())
            if self.debug:
                return pdb.runcall(self.handleNextInput)

            else:
                return self.handleNextInput()

        except EOFError as e:
            print()

            e = str(e)
            if e:
                print(f'{self.__class__.__name__}.handleConsoleInput: {e}')

            from stuphos.etc.tools.strings import isYesValue
            if isYesValue(configuration.Interpreter.console_eof_stubborn):
                # If not set, seek to terminate this console session.
                return

        except SystemExit:
            pass
        else:
            return

        # Full stop on errors.
        if self.engine.headless:
            self.engine -= self
        else:
            @self.engine.event.call
            def done():
                # debugOn()
                self.engine.stop()
                StuphMUD.ShutdownGame()
                # gc.collect(); os.kill(signal.SIGTERM, os.getpid())


    # Entity Handle -- Bastard player implementation.
    def __init_entity__(self, debug = False):
        self.host = 'localhost'
        self.debug = debug

    def textout(self, text):
        # Set by mud.player.shell.ShellI.withPeerHeadAndException
        getattr(sys, '__peer_shell_stdout').write(text)

    def sendln(self, line):
        print(line, file=sys.stdout)

    # Avoid the whole external pipe thing, especially for a windowed IDE.
    if os.name.startswith('nt'):
        def page_string(self, string):
            from stuphos.etc.tools.strings import parse_color
            string = ''.join(parse_color(string))
            print(string)

    elif os.name.startswith('posix'):
        nroff = True
        def page_string(self, string):
            from os import popen
            from stuphos.etc.tools.strings import parse_color
            string = ''.join(parse_color(string))
            popen('%s/bin/less' % ('/usr/bin/nroff|' if self.nroff else ''), 'w') \
                                .write(str(string))

    def editString(self, string):
        if callable(self.messenger):
            self.messenger(self, string)

    def messenger(self, peer, string):
        self.page_string(string)




import world

from world.player import playerByID
from world.player.nanny import CON_GET_NAME

    # players = world.player.players
    # zone = world.zone
    # room = world.room
    # item = world.item
    # mobile = world.mobile
    # iterate_entities = world.iterate_entities

# from stuphos.runtime.facilities import Facility

try: import game
except ImportError:
    # Copy of framework/game.py
    from stuphos.system.game import *

    import world

    from world import peer, mobile_instance, player_character, getPlayerStoreByName, iteratePlayerIDs, findPlayerNameByID, SPECIAL
    from world.player import getMasterCommandList, interpretCommand, loadPlayer, EnableConsole
    from world.player import nanny, commands
    from world.player.commands import lookAtRoom

else:
    from .game import mudlog, syslog, iterateOLCAccessByName



# Supposed to replace Circle board functionality, and provide a place to extend it.
# For now, it seeks to emulate Circle as closely as possible.
# classic

# expansion:
#   stylus-items for indexing deep into message thread-tree
#   for an hierarchical discussion board.
#
# Otherwise, the BOARD protocol should provide indirection
# layer between classical and this module models.

from time import time as now, asctime, localtime
from datetime import datetime

from stuphmud.server.tools import splitOne as chopOne, SORP
from stuphmud.server.runtime import Object

SECONDS_PER_DAY = 60 * 60 * 24

# System.
def getCurrentDateTime():
    return datetime.fromtimestamp(now())

def getTimeString(time):
    if type(time) is datetime:
        return time.ctime()
        time = time

    return asctime(localtime(time))

# Tools.
def delete_doubledollar(string):
    return string.replace('$$', '$')

def parseSpecialCommand(cmd, argstr):
    cmd = getattr(cmd, 'name', cmd)
    assert type(cmd) is str, TypeError(type(cmd).__name__)

    args = argstr and argstr.split() or ()
    return (cmd, args)

def isName(name, namelist):
    return name.lower() in namelist.lower().split()

# BOARD Model Implementation.
class Board(Object):
    _Meta = Object._Meta('name', 'vnum', 'filename')

    ACT_BEGIN_WRITING_TOCHAR = "Write your message.  End with '.' or '@' on a new line.\n"
    ACT_BEGIN_WRITING_TOROOM = "$n starts to write a message."

    MSG_NO_WRITE_TOCHAR = 'You are not holy enough to write on this board.'
    MSG_NO_READ_TOCHAR = 'You try but fail to understand the holy words.'

    MSG_NOTHOLY_ENOUGH = "You are not holy enough to remove other people's messages."
    MSG_NOTHOLIER_THAN = "You can't remove a message holier than yourself."

    class Message(Object):
        _Meta = Object._Meta('headline', 'author_name')

        def __init__(self, board, headline = '',
                     timestamp = None,
                     author_idnum = -1,
                     author_level = 0,
                     author_name = ''):

            self.board = board
            self.headline = headline
            self.timestamp = timestamp

            self.author_idnum = author_idnum
            self.author_level = author_level
            self.author_name = author_name

            self.body = ''

        # Creation.
        def startWriting(self, author):
            startWriting(author, self.finishWriting,
                         self.board.ACT_BEGIN_WRITING_TOCHAR,
                         self.board.ACT_BEGIN_WRITING_TOROOM)

        def finishWriting(self, author, string):
            self.body = string
            self.board.saveBoard()

        # Evaluation.
        def getAgeInDays(self):
            age = (getCurrentDateTime() - self.timestamp).days
            return ' (%s day%s old)' % (age, SORP(age))

        def getFullHeadline(self, withAge = False):
            return '&G%6.10s &C%-12s &N:: &b%s&N%s' % \
                   (getTimeString(self.timestamp),
                    '(%s)' % self.author_name, self.headline,
                    self.getAgeInDays() if withAge else '')

        def getFullDisplay(self, nr):
            return 'Message %d : %s\n\n%s\n' % (nr, self.getFullHeadline(), self.body)

        def isAuthor(self, author):
            return self.author_idnum == author.idnum and \
                   self.author_name == author.name

        def canRemove(self, actor, explain_why = False):
            return self.board.canRemove(actor, self, explain_why = explain_why)
        def remove(self, actor, nr):
            self.board.removeMessage(actor, self, nr)

    @classmethod
    def FromStuphLib(self):
        from stuphlib.boards import _builtin_boards as builtin
        return list(map(self.FromBuiltin, builtin))

    @classmethod
    def FromBuiltin(self, builtin):
        # Create a board object from a builtin record.
        return self(vnum = builtin.vnum,
                    name = builtin.name,
                    filename = builtin.name,
                    read_level = builtin.read_lvl,
                    write_level = builtin.write_lvl,
                    remove_level = builtin.remove_lvl,
                    auto_cleanup = builtin.auto_cleanup)

    def __init__(self, vnum = None, name = None, filename = None,
                 read_level = None, write_level = None, remove_level = None,
                 auto_cleanup = None):

        self.vnum = vnum
        self.name = name
        self.filename = filename
        self.read_level = read_level
        self.write_level = write_level
        self.remove_level = remove_level
        self.auto_cleanup = auto_cleanup

        self.messages = []

    def saveBoard(self):
        if self.filename:
            import json
            def ofMessage(msg):
                return dict(headline = msg.headline,
                            timestamp = msg.timestamp.ctime(),

                            author_idnum = msg.author_idnum,
                            author_level = msg.author_level,
                            author_name = msg.author_name,

                            body = msg.body)

            with open(self.filename, 'w') as o:
                json.dump(dict(vnum = self.vnum,
                               name = self.name,
                               read_level = self.read_level,
                               write_level = self.write_level,
                               remove_level = self.remove_level,
                               auto_cleanup = self.auto_cleanup,
                               messages = list(map(ofMessage, self.messages))),
                          o, indent = 1)

    def canWrite(self, author, explain_why = False):
        if author.level >= self.write_level:
            return True
        if explain_why and author.peer:
            print(self.MSG_NO_WRITE_TOCHAR, file=author.peer)

        return False

    def canRead(self, author, explain_why = False):
        if author.level >= self.read_level:
            return True
        if explain_why and author.peer:
            print(self.MSG_NO_READ_TOCHAR, file=author.peer)

        return False

    def canRemove(self, author, msg, explain_why = False):
        if author.level < self.remove_level and not msg.isAuthor(author):
            if explain_why and author.peer:
                print(self.MSG_NOTHOLY_ENOUGH, file=author.peer)

            return False

        if author.level < msg.author_level and not author.supreme:
            if explain_why and author.peer:
                print(self.MSG_NOTHOLIER_THAN, file=author.peer)
            return False

        return True

    def createMessage(self, author, headline):
        msg = self.Message(self, headline,
                           timestamp = getCurrentDateTime(),
                           author_idnum = author.idnum,
                           author_level = author.level,
                           author_name = author.name)

        self.messages.append(msg)
        return msg

    def removeMessage(self, author, message, nr):
        self.messages.remove(message)
        if author.peer:
            print('Message removed.', file=author.peer)

        # author.act('$n just removed message %d.' % nr)
        self.saveBoard()

    def registerItem(self, item):
        item.__board__ = self
        item.special = specialBoard

class BoardManager:
    def __init__(self):
        self.table = {}
    def __getitem__(self, item):
        try: return self.table.get(item.vnum)
        except ValueError:
            # Atypical.
            raise KeyError

_builtin_board_manager = BoardManager()

_builtin_boards = Board.FromStuphLib()

_builtinBoardsByName = dict((b.name, b) for b in _builtin_boards)
_builtinBoardsByVnum = dict((b.vnum, b) for b in _builtin_boards)


# Worldly Actualization.
def findBoardByItem(item):
    try: return item.__board__
    except AttributeError:
        pass

    try: return item.prototype.__board__
    except AttributeError:
        pass

    try: return _builtin_board_manager[item]
    except KeyError:
        pass

def findNamedBoard(actor, name, this):
    # From boards.cpp:find_board(ch, arg)

    # "If there's an argument we look by keyword first"
    name = name.strip()
    if name:
        # "First check inventory by name"
        item = actor.find(name, inventory = True)
        if item is not None:
            board = findBoardByItem(item)
            if board is not None:
                return (item, board)

        # "Didn't find that lets try the room"
        item = actor.find(name, item_in_room = True)
        if item is not None:
            board = findBoardByItem(item)
            if board is not None:
                return (item, board)

    # "Ok we'll resort to this..."
    for item in actor.room.contents:
        board = findBoardByItem(item)
        if board is not None:
            return (item, board)

    for item in actor.inventory:
        board = findBoardByItem(item)
        if board is not None:
            return (item, board)

    # Default result (this object).
    board = findBoardByItem(this)
    if board is not None:
        return (this, board)

    return (None, None)

def startWriting(author, finish, to_char, to_room):
    peer = author.peer
    if peer:
        def messageTerminator(peer, string):
            author.playerflags.writing = False
            finish(author, string)

        peer.messenger = messageTerminator

        if not author.npc:
            author.playerflags.writing = True

        print(to_char, file=peer)
        peer.editString('')

        # XXX
        # author.act(to_room, hide_if_invisible = True, to_room = True)

# Player Interface.
# Todo: what about NEWEST_AT_TOP emulation?  Newest is always at bottom..
def doReadBoard(actor, this, argstr, *args):
    # Peer is required for message output.
    if not actor.peer:
        return False

    # Not supporting "read board" currently..

    (name, argstr) = chopOne(argstr, ' ')
    (item, board) = findNamedBoard(actor, name, this)
    if board is not None and argstr.isdigit(): # read 2.mail, look 2.sword
        if board.canRead(actor, explain_why = True):
            msgNr = int(argstr)

            if not board.messages:
                print('The board is empty!', file=actor.peer)
            elif msgNr < 1 or msgNr > len(board.messages):
                print('That message exists only in your imagination!', file=actor.peer)
            else:
                msg = board.messages[msgNr - 1]
                if not msg.body:
                    print('That message seems to be empty.', file=actor.peer)
                else:
                    actor.peer.page(msg.getFullDisplay(msgNr))

        return True

def doWriteBoard(actor, this, argstr, *args):
    # Peer is required for message input.
    if not actor.peer:
        return False

    (name, argstr) = chopOne(argstr, ' ')
    (item, board) = findNamedBoard(actor, name, this)
    if board is not None:
        if board.canWrite(actor, explain_why = True):
            argstr = argstr.lstrip()
            argstr = delete_doubledollar(argstr)

            if len(argstr) > 80:
                argstr = argstr[:80]

            if not argstr:
                print('We must have a headline!', file=actor.peer)
            else:
                msg = board.createMessage(actor, argstr)
                msg.startWriting(actor)

        return True

def doShowBoard(actor, this, argstr, *args):
    # Peer is required for message display.
    if not actor.peer:
        return False
    if not args or not isName(args[0], this.keywords):
        return False

    (item, board) = findNamedBoard(actor, argstr, this)
    if board is not None:
        if board.canRead(actor, explain_why = True):
            # XXX
            # actor.act('$n studies %s.' % item.name,
            #           hide_if_invisible = True, to_room = True)

            # Display
            def _():
                n = len(board.messages)
                yield '&NThis is %s.  &NUsage: &cREAD/REMOVE&N <messg #>, &cWRITE&N <header>.' % item.name

                # This may not be true, but it may prompt other players to look at the board.
                yield 'You will need to look at the board to save your message.'

                if not n:
                    yield 'The board is empty.'
                else:
                    yield 'There are %d message%s on the board.' % (n, SORP(n))

                    for (i, m) in enumerate(board.messages):
                        yield '%-3d: %s' % (i + 1, m.getFullHeadline(withAge = True))

                yield ''

            actor.peer.page('\n'.join(_()))

        return True

def doRemoveFromBoard(actor, this, argstr, *args):
    def notify(msg):
        if actor.peer:
            print(msg, file=actor.peer)

    (name, argstr) = chopOne(argstr, ' ')
    (item, board) = findNamedBoard(actor, name, this)
    if board is not None and argstr.isdigit():
        from stuphmud.server.player import playerAlert, PlayerAlert
        try:
            n = len(board.messages)
            if not n:
                playerAlert('The board is empty!')
            else:
                msgNr = int(argstr)
                if msgNr < 1 or msgNr > n:
                    playerAlert('That message exists only in your imagination.')
                else:
                    msg = board.messages[msgNr - 1]
                    if msg.canRemove(actor, explain_why = True):
                        # Todo: 'At least wait until the author is finished before removing it!'
                        msg.remove(actor, msgNr)

        except PlayerAlert as e:
            if actor.peer:
                e.deliver(actor.peer)

        return True

#@ASSIGNOBJ('BOARD::Interface')
def specialBoard(actor, this, cmd, argstr):
    (cmd, args) = parseSpecialCommand(cmd, argstr)
    func = _board_functions.get(cmd)

    # Originally, the board_type is found before calling the command,
    # but this leaves that specific argument-matching up to that routine.
    if callable(func):
        return func(actor, this, argstr, *args)

_board_functions = dict(write   = doWriteBoard,
                        read    = doReadBoard,
                        look    = doShowBoard,
                        examine = doShowBoard,
                        remove  = doRemoveFromBoard)


# This is a departure from Circle, which would define boards internally.
_board_cache = {}

from stuphmud.server.player import ACMDLN, Option
@ACMDLN('new-board', Option('--board-name', '--name'),
                     Option('--item'),
                     Option('--vnum', type = int),
                     Option('--filename'),
                     Option('--read-level', type = int),
                     Option('--write-level', type = int),
                     Option('--remove-level', type = int),
                     Option('--auto-cleanup', action = 'store_true'))
def doNewBoard(peer, command):
    player = peer.avatar
    if player.supreme:
        boardName = command.options.board_name
        itemName = command.options.item

        if not (boardName and itemName):
            yield command.help()
        else:
            try: board = _board_cache[boardName]
            except KeyError:
                # Figure out what this item is.
                if itemName.isdigit():
                    import world
                    item = world.item(int(itemName))

                else:
                    item = player.find(itemName, inventory = True, equipment = True,
                                       item_in_room = True)

                if item is None:
                    yield 'Could not find any item named %r' % itemName
                else:
                    # Create new board, add it to cache, register item.
                    o = command.options
                    board = Board(name = boardName,
                                  vnum = o.vnum,
                                  filename = o.filename,
                                  read_level = o.read_level,
                                  write_level = o.write_level,
                                  remove_level = o.remove_level,
                                  auto_cleanup = o.auto_cleanup)

                    _board_cache[boardName] = board
                    board.registerItem(item)

                    yield 'New board created [%s], and attached it to %s.' % (boardName, item)
            else:
                yield 'Board already exists: %r' % board



# Provides a common configuration source for the xmlrpc host/client.
CONFIG_FILE = ''

class DEFAULT_CONFIG:
    BIND_HOST = 'localhost' # '0.0.0.0'
    PORT = 2172

    hostname = BIND_HOST
    port = PORT

class NotConfigured(RuntimeError):
    pass

def buildConfig(config = None):
    from stuphos.etc.tools import isYesValue
    from stuphos import getConfigObject
    xmlrpc = getConfigObject().getSection('XMLRPC')

    if isYesValue(xmlrpc['off']):
        raise NotConfigured

    cfg = DEFAULT_CONFIG()

    address = xmlrpc['address']
    if address is not None:
        cfg.hostname = address

    port = xmlrpc['port']
    if port is not None:
        cfg.port = int(port)

    cfg.certificate = xmlrpc.get('certificate-path', None)
    cfg.keyfile = xmlrpc.get('keyfile-path', None)

    return cfg



# Routines for the StuphMUD Text Library.
from os.path import join as joinpath
from os.path import basename, splitext

def ReadTildeString(fp, area = 'no particular', newline = '\n'):
    '''
    Read lines from file until terminated by "~\n", removing this sequence.
    Return 2-tuple of (text, nr_lines).

    '''

    buffer = []
    append = buffer.append

    nrlines = 0
    while True:
        line = fp.readline()
        nrlines += 1

        if line == '':
            raise EOFError('Area: %s [EOF expecting ~string]' % area)

        # XXX should not strip every line (is this done in C source??)
        line = line.strip()
        tildeNdx = line.find('~')
        last = len(line) - 1

        if tildeNdx > -1:
            if tildeNdx == last and tildeNdx != 0:
                line = line[:-1]
                append(line)

            break

        else:
            append(line)

    # There is a tilde: put it on, break on ~\n
    # Q: What about ending newline?
    buffer = newline.join(buffer)

    return (buffer, nrlines)

def ReadTextFile(fp):
    # like, most found in the "lib/text" directory"
    return fp.read()
ReadMiscFile = ReadTextFile

def ReadFileNumber(fp):
    line = fp.readline()
    if line == '':
        raise EOFError

    return int(line.strip())

def ReadFightMessages(fp, none_value = ''):
    '''
    * Note: all lines between records which start with '*' are comments and
    * are ignored.  Comments can only be between records, not within them.
    *
    * This file is where the damage messages go for offensive spells, and
    * skills such as kick and backstab.  Also, starting with Circle 3.0, these
    * messages are used for failed spells, and weapon messages for misses and
    * death blows.
    *
    * All records must start with 'M' in the first column (for 'Message').
    * The next line must contain the damage number (defined in spells.h),
    * then the messages (one per line):
    *   Death Message (damager, damagee, onlookers)
    *   Miss Message (damager, damagee, onlookers)
    *   Hit message (damager, damagee, onlookers)
    *   God message (damager, damagee, onlookers)
    *
    * All messages must be contained in one line.  They will be sent through
    * act(), meaning that all standard act() codes apply, and each message will
    * automatically be wrapped to 79 columns when they are printed.  '#' can
    * be substituted for a message if none should be printed.  Note however
    * that, unlike the socials file, all twelve lines must be present for each
    * record, regardless of any '#' fields which may be contained in it.
    '''

    linenr = [0] # as cell
    def readMessageLine(fp, message = 'fight'):
        line = fp.readline()
        if line == '':
            raise SyntaxError('Expected %s message (line %d)' % (message, linenr[0]))

        linenr[0] += 1
        if line.strip() == '#':
            return none_value

        # Q: Strip newline?  "All messages must be contained in one line."
        return line

    last_comment = none_value
    while True:
        # Read up to the first part of the message record.
        line = fp.readline()
        if line == '':
            # Expected a terminating '$', but...
            break

        linenr[0] += 1
        line = line.strip()
        if not line:
            continue

        if line[0] == '*':
            last_comment = line[1:].strip()
            continue

        # Beginning of message record.
        if line == '$':
            # The only proper way to terminate the file.
            break

        if line != 'M':
            raise SyntaxError('Expected message marker ("M"), got %r (line %d)' % (line, linenr[0]))

        line = fp.readline()
        if line == '':
            raise SyntaxError('Expected message code (line %d)' % linenr[0])

        start_line = linenr[0]
        linenr[0] += 1
        line = line.strip()
        if not line.isdigit():
            raise SyntaxError('Expected message code, got %r (line %d)' % (line, linenr[0]))

        code = int(line) # id

        # Now read messages.
        yield dict(message_code = code,
                   line_number = start_line,
                   last_comment = last_comment,

                   death = dict(damager = readMessageLine(fp, 'death-damager'),
                                damagee = readMessageLine(fp, 'death-damagee'),
                                onlookers = readMessageLine(fp, 'death-onlookers')),
                   miss = dict(damager = readMessageLine(fp, 'miss-damager'),
                               damagee = readMessageLine(fp, 'miss-damagee'),
                               onlookers = readMessageLine(fp, 'miss-onlookers')),
                   hit = dict(damager = readMessageLine(fp, 'hit-damager'),
                              damagee = readMessageLine(fp, 'hit-damagee'),
                              onlookers = readMessageLine(fp, 'hit-onlookers')),
                   god = dict(damager = readMessageLine(fp, 'god-damager'),
                              damagee = readMessageLine(fp, 'god-damagee'),
                              onlookers = readMessageLine(fp, 'god-onlookers')))

        last_comment = none_value

class FightMessage:
    class MessageSet:
        def __init__(self, kwd):
            self.__dict__.update(kwd)

        def __repr__(self):
            return 'damager = %r\ndamagee = %r\nonlookers = %r' % \
                   (self.damager, self.damagee, self.onlookers)

    def __init__(self, message_code = None, line_number = None, last_comment = None,
                 death = None, miss = None, hit = None, god = None):

        self.message_code = message_code
        self.line_number = line_number
        self.last_comment = last_comment

        self.death = self.MessageSet(death)
        self.miss  = self.MessageSet(miss)
        self.hit   = self.MessageSet(hit)
        self.god   = self.MessageSet(god)

    def __repr__(self):
        return ('%s:\n  code: #%d\n  line: #%d\n  comment: %r\n' + \
                'death: %s\n  miss : %s\n  hit  : %s\n  god  : %s\n') % \
               (self.__class__.__name__, self.message_code,
                self.line_number, self.last_comment,
                self.death, self.miss, self.hit, self.god)

def ReadSocialMessages(fp):
    # Todo: incorporate line-number information into errors
    linenr = [0] # as cell

    def readSocialLine():
        line = fp.readline()
        if line == '':
            raise EOFError

        linenr[0] += 1
        line = line.strip()
        return line

    while True:
        try: line = readSocialLine()
        except EOFError:
            # Expected a terminating '$', but...
            break

        if line == '$':
            # The only proper termination:
            break

        # Todo: interpret and re-raise error info for this operation:
        (social_name, hide, position) = line.split()

        social = dict(social_name = social_name.lower(),
                      hide = hide,
                      position = position,
                      line_number = linenr[0])

        # Todo: explicitly blank out non-specified lines?

        # Start of messages: end with blank line.
        # Pay close attention to the role of '#'
        line = readSocialLine()
        social['char_no_arg'] = line

        line = readSocialLine()
        if line != '#':
            social['others_no_arg'] = line

        line = readSocialLine()
        if line != '#':
            social['char_found'] = line
            line = readSocialLine()
            social['others_found'] = line
            line = readSocialLine()
            social['vict_found'] = line
            line = readSocialLine()
            social['not_found'] = line
            line = readSocialLine()
            social['char_auto'] = line
            line = readSocialLine()
            if line != '#':
                social['others_auto'] = line

        # End of record marker (empty-line)
        line = readSocialLine()
        if line:
            if not social_name:
                raise SyntaxError('Missing end of record near beginning of file')
            else:
                raise SyntaxError('Missing end of record marker near social %r' % social_name)

        yield social

class SocialMessage:
    def __init__(self, social_name = None, hide = None, position = None,
                 line_number = None, char_no_arg = None, others_no_arg = None,
                 char_found = None, others_found = None, vict_found = None,
                 not_found = None, char_auto = None, others_auto = None):

        self.social_name = social_name
        self.hide = hide
        self.position = position
        self.line_number = line_number
        self.char_no_arg = char_no_arg
        self.others_no_arg = others_no_arg
        self.char_found = char_found
        self.others_found = others_found
        self.vict_found = vict_found
        self.not_found = not_found
        self.char_auto = char_auto
        self.others_auto = others_auto

    def __repr__(self):
        return ('%s: %r (%r; %s; line #%d)\n' + \
                '  char_no_arg = %r\n' + \
                '  others_no_arg = %r\n' + \
                '  char_found = %r\n' + \
                '  others_found = %r\n' + \
                '  vict_found = %r\n' + \
                '  not_found = %r\n' + \
                '  char_auto = %r\n' + \
                '  others_auto = %r\n') % \
                (self.__class__.__name__,
                 self.social_name, bool(self.hide), self.position,
                 self.line_number,
                 self.char_no_arg, self.others_no_arg,
                 self.char_found, self.others_found,
                 self.vict_found, self.not_found,
                 self.char_auto, self.others_auto)

def ReadOlcZones(fp):
    linenr = 0
    while True:
        line = fp.readline()
        if line == '':
            # Expected a terminating '$', but...
            break

        line = line.strip()
        linenr += 1
        if line == '$':
            # The only proper termination:
            break

        # Read zone vnum.
        try:
            if not line or line[0] != '#':
                raise ValueError

            vnum = int(line[1:])

        except ValueError:
            raise SyntaxError('Expected #<zone vnum>, got %r (line #%d)' % (line, linenr))

        # Read player idnums.
        idnums = []
        while True:
            line = fp.readline()
            if line == '':
                raise EOFError('Expected list of idnums (line #%d)' % line)

            line = line.strip()
            linenr += 1

            if line == 'S':
                # Normal end of idnums.
                break

            try: idnums.append(int(line))
            except ValueError:
                raise SyntaxError('Expected idnum, got %r (line #%d)' % (line, linenr))

        # Return result.
        yield (vnum, idnums)

def ReadTrustFile(fp):
    linenr = 0
    while True:
        line = fp.readline()
        if line == '':
            # Expected a terminating '$', but...
            break

        line = line.strip()
        linenr += 1
        if line == '$':
            # The only proper termination:
            break

        # Read player idnum.
        try:
            if not line or line[0] != '#':
                raise ValueError

            idnum = int(line[1:])

        except ValueError:
            raise SyntaxError('Expected #<player idnum>, got %r (line #%d)' % (line, linenr))

        # Read command names.
        commands = []
        while True:
            line = fp.readline()
            if line == '':
                raise EOFError('Expected list of command names (line #%d)' % line)

            line = line.strip()
            linenr += 1

            if line == '*':
                # Normal end of commands.
                break

            if not line:
                raise SyntaxError('Expected command name or *, got empty line (line #%d)' % line)

            commands.append(line)

        # Return result.
        yield (idnum, commands)

def ReadMudTime(fp):
    return ReadFileNumber(fp)

def ReadDnsCache(fp):
    linenr = 0
    while True:
        line = fp.readline()
        if line == '':
            # End of Records.
            break

        linenr += 1
        line = line.strip()
        if not line:
            raise SyntaxError('Expected <ip> <hostname>, got empty line (line #%d)' % linenr)

        # Todo: interpret and re-raise error info for this operation:
        (ipaddr, hostname) = line.split()

        # Todo: further process ipaddr x.x.x.x into host-order int value.
        yield (ipaddr, hostname)

def ReadXNames(fp):
    return set([_f for _f in fp if _f])

_misc_files = ['bugs', 'ideas', 'typos']
_text_files = ['background', 'clanhelp', 'credits', 'handbook',
               'immlist', 'imotd', 'info', 'motd', 'news', 'policies',
               'title', 'wizlist']

# Player Files.
_middle_names = ['A-E', 'F-J', 'K-O', 'P-T', 'U-Z', 'ZZZ']
_middle_name_initials = [(ord(m[0]), ord(m[2])) for m in _middle_names[:-1]]

def GetMiddleInitial(name):
    try: i = ord(name[0].upper())
    except IndexError:
        raise ValueError(name)

    for x in range(len(_middle_name_initials)):
        range = _middle_name_initials[x]
        if i >= range[0] and i <= range[1]:
            return _middle_names[x]

    return _middle_names[-1]

ALIAS_SIMPLE	= 0
ALIAS_COMPLEX	= 1

ALIAS_SEP_CHAR	= ';'
ALIAS_VAR_CHAR	= '$'
ALIAS_GLOB_CHAR	= '*'

NUM_TOKENS	    = 9

def ReadPlayerAliases(fp):
    linenr = 0
    try:
        while True:
            (alias, nrlines) = ReadTildeString(fp, area = '')

            linenr += nrlines

            (replacement, nrlines) = ReadTildeString(fp, area = 'alias replacement')
            linenr += nrlines

            line = fp.readline()
            if line == '':
                raise EOFError('Expected alias complexity type (line #%d)' % linenr)

            linenr += 1
            line = line.strip()
            try: complex = int(line)
            except ValueError:
                raise SyntaxError('Expected alias complexity type, got %r (line #%d)' % (line, linenr))

            yield dict(alias       = alias,
                       replacement = replacement,
                       type        = complex)
    except EOFError:
        # Normal end of records.
        pass

class Alias:
    def __init__(self, alias = None, replacement = None, type = None):
        self.alias       = alias
        self.replacement = replacement
        self.type        = type

    def IsComplex(self):
        return self.type == ALIAS_COMPLEX

    def __repr__(self):
        return '%s%s: %s => %r' % (self.__class__.__name__,
                                   self.IsComplex() and ' (Complex)' or '',
                                   self.alias, self.replacement)

# Order in file:
_player_strings = ('title', 'prename', 'wizname',
                   'poofin', 'poofout',
                   'description', 'plan')

def ReadPlayerStrings(fp):
    # ignoring line number information
    return dict((name, ReadTildeString(fp, 'player ' + name)[0]) \
                for name in _player_strings)

class PlayerStrings:
    def __init__(self, title = None, prename = None, wizname = None,
                 poofin = None, poofout = None, description = None, plan = None):

        self.title       = title
        self.prename     = prename
        self.wizname     = wizname
        self.poofin      = poofin
        self.poofout     = poofout
        self.description = description
        self.plan        = plan

    def __repr__(self):
        return 'Strings:\n  %s' % \
               '\n  '.join('%s = %r' % (name, getattr(self, name)) \
                           for name in _player_strings)

# Mail.
def ReadMail(fp):
    # Todo: parse header info?  Use rfc822.Message?
    try:
        while True:
            (message, nrlines) = ReadTildeString(fp, 'player mail message')
            yield message

    except EOFError:
        # Normal end of records.
        pass

def ReadVerifiedMail(directory, name, middle = None):
    # Pass lib/plrmail directory and player name.
    name = name.lower()
    if middle is False:
        filebase = joinpath(directory, name)
    else:
        if middle is None:
            middle = GetMiddleInitial(name)

        filebase = joinpath(directory, middle, name)

    # Verify number of messages
    count = ReadFileNumber(open('%s.count' % filebase))
    mail = ReadMail(open('%s.mail' % filebase))
    mail = list(mail)

    if len(mail) != count:
        # XXX: Yes, there's a format error, but could we just truncate the
        # rest?  Or just pass the count to the ReadMail subroutine to limit?
        raise ValueError('Expected %d messages, got %d' % (count, len(mail)))

    return mail

# Inspection Front End.
def _get_interpreters():
    def KwdObjectList(array, objectClass):
        return [objectClass(**kwd) for kwd in array]

    interpreters = {'messages' : (ReadFightMessages,
                                 lambda kwd: KwdObjectList(kwd, FightMessage)),
                    'socials'  : (ReadSocialMessages,
                                 lambda kwd: KwdObjectList(kwd, SocialMessage)),

                    'olczones' : (ReadOlcZones , dict),
                    'trust'    : (ReadTrustFile, dict),
                    'time'     : (ReadMudTime  , int ),
                    'dns.cache': (ReadDnsCache , dict),

                    # Extensions.
                    '.alias'   : (ReadPlayerAliases,
                                  lambda kwd: KwdObjectList(kwd, PlayerAlias)),
                    '.strings' : (ReadPlayerStrings,
                                  lambda kwd: PlayerStrings(**kwd)),

                    # Directories.
                    'plralias'  : (lambda *args, **kwd: [],
                                   lambda n: None),
                    'plrstrings': (lambda *args, **kwd: [],
                                   lambda n: None),
                    'plrmail'   : (lambda *args, **kwd: [],
                                   lambda n: None)}

    for filename in _misc_files:
        interpreters[filename] = (ReadTextFile, str)
    for filename in _text_files:
        interpreters[filename] = (ReadTextFile, str)

    return interpreters

def _lookup_interpreter(interpreters, filename):
    filename = basename(filename).lower()
    try: return interpreters[filename]
    except KeyError:
        ext = (splitext(filename)[1] or '').lower()
        return interpreters.get(ext)

def main(argv = None):
    from optparse import OptionParser
    from code import InteractiveConsole as IC
    from types import InstanceType
    from pdb import runcall
    import readline

    try: from simplejson import dump
    except ImportError:
        from pprint import pprint as output
    else:
        from sys import stdout
        def output(record):
            dump(record, stdout, indent = 2)
            print(',', file=stdout)

    parser = OptionParser()
    parser.add_option('-i', '--inspect', action = 'store_true')
    (options, args) = parser.parse_args(argv)

    # Build these.
    interpreters = _get_interpreters()

    for filename in args:
        i = _lookup_interpreter(interpreters, filename)
        if i is not None:
            (loader, internalize) = i
            object = loader(open(filename))
            object = internalize(object)

            if options.inspect:
                object_name = basename(filename).replace('.', '_')
                IC(locals = {object_name: object}).interact(banner = '')
            elif type(object) is InstanceType:
                print(object)
            else:
                # Todo: control formatted output options.
                output(object)

if __name__ == '__main__':
    main()



from . import CommandMessage, RegisterCommandClass
from ...etc.hashlib import md5
from ...etc.timing import getCurrentSystemTime


def md5sum(*parts):
    hash = md5()
    for p in parts:
        hash.update(p)

    return hash.hexdigest()

class MaterializeResourceCommandMessage(CommandMessage):
    # Memcache equivilant?
    class Pool(dict):
        @classmethod
        def createMaster(self):
            master = self(':MASTER:') # Kind of a container of all pools.
            master.default = self('')
            return master

        RESOURCE_POOL_MASTER_NAME = 'Resource::Pool::Master'

        @classmethod
        def getMaster(self):
            from ....runtime.registry import getObject
            return getObject(self.RESOURCE_POOL_MASTER_NAME,
                             create = self.createMaster)

        def __init__(self, name):
            dict.__init__(self)
            self.name = name

        def addResource(self, name, rsrc):
            self.setdefault(name, []).append(rsrc)
        def getByNamedId(self, name, id = None):
            listed = self.get(name)
            if id is None:
                return listed # or first

            for rsrc in listed:
                if rsrc.id == id:
                    return rsrc

            # Raise??

        def __repr__(self):
            return '<%s: %d entries>' % (self.name, len(self))

    class Resource:
        def __init__(self, id, content_type, payload, original = None, pool = None):
            self.id = id
            self.content_type = content_type
            self.payload = payload
            self.original = original
            self.pool = pool

        def __repr__(self):
            pool = repr(self.pool) if self.pool is not None else ''
            return '<%s: %r%s%s>' % (self.__class__.__name__, self.id,
                                     pool and ' ' or '', pool)

        def __str__(self):
            return str(self.payload)

    @classmethod
    def getPool(self, name):
        master = self.Pool.getMaster()
        if name is None:
            return master.default

        try: return master[name]
        except KeyError:
            pool = master[name] = self.Pool(name)
            return pool

    @classmethod
    def getMessageId(self, pool_name, name, headers, payload):
        try: id = headers['X-Resource-Id']
        except KeyError:
            id = md5sum(getCurrentSystemTime(), payload)

        return '%s:%s:%s' % (pool_name, name, id)

    @classmethod
    def getResource(self, id = None, pool_name = None, name = None):
        if pool_name:
            assert name
        else:
            assert id
            assert not name

            parts = id.split(':', 3)
            if len(parts) == 3:
                (pool_name, name, id) = parts
            elif len(parts) == 2:
                (pool_name, name) = parts
                id = None
            else:
                raise AssertionError('Resource id format error: %s' % id)

        pool = self.getPool(pool_name)
        return pool.getByNamedId(name, id)

    def Execute(self):
        # Yet another embedded message??
        # The (resource) message content and headers should be merged with
        # the communication message, avoiding this extra embedding level.
        # msg = message_from_string(self.payload)

        pool_name = self['X-Resource-Pool-Name']
        pool = self.getPool(pool_name)

        name = msg.get_filename() # XXX err, resolve this from payload?
        id = self.getMessageId(pool_name, name, self.getHeaders(), self.payload)

        pool.addResource(name,
                         self.Resource(id,
                                       self.getContentType(),
                                       self.getHeaders(),
                                       self.payload))

RegisterCommandClass('application/x-stuph-material-resource', MaterializeResourceCommandMessage)



# federization module
# move this into mud/runtime?
def open(frame, host, port, username, password, api):
	if frame.task.checkAccess(['system/remote', '%s:%s' % (host, port), api]):
		from receptacle.client.support import Authorization, APICall
		client = Authorization(username, password).Open(host, port)

		def remoteApiCallCommand(*args, **kwd):
			# todo: asynchronously
			return client.callCommand(*args, **kwd)
		def remoteApiGetEntityProperty(*args, **kwd):
			# todo: asynchronously
			return client.getEntityProperty(*args, **kwd)

		apiCall = APICall(remoteApiCallCommand, remoteApiGetEntityProperty)
		return apiCall.open(api).invoke


# todo: structural item that configures host/port api
# and connects auth (account : {username: password})
# provides subnative methodological api

def installApi(frame, name, configObject):
	# install service api or application mode.
	# check access, invoke with configuration for core.
	pass

@apply
class network:
	class group:
		def __init__(self, frame, name):
			# Return a broadcaster by looking up databased directory network group.
			pass



#!/usr/bin/python
# Copyright 2021 runphase.com .  All rights reserved.
#
import os
import sys
import pdb
import queue

from os.path import realpath, normpath, dirname

# from .system import *


# Front-End.
def buildCli(consoleBaseClass = None):
    # Do this dynamically -- import the core stuff only during core compartmentalize.
    global Console, Heartbeat, peer, mobile_instance, NanoText
    from game import Heartbeat, peer, mobile_instance, NanoText # XXX remove reliance on game

    import sys
    sys.__peer_shell_stdout = sys.stdout

    if consoleBaseClass is None:
        from stuphos.system.game.compartment.system import ConsoleBusiness as consoleBaseClass

    class Console(consoleBaseClass, peer, Heartbeat.Task):
        def __init_entity__(self, *args, **kwd):
            peer.__init_entity__(self) # conn = something containing network ref
            consoleBaseClass.__init_entity__(self, *args, **kwd)
            self.avatar = mobile_instance.create()

        def editString(self, string):
            string = NanoText(string)
            return consoleBaseClass.editString(self, string)

        class Nonconnection:
            def __init__(self, peer, network):
                self.peer = peer
                self.network = network

        def attachNetwork(self, network):
            self.conn = self.Nonconnection(self, network)

# Executable Script.
# ['mushclient_game_cli', '-m', 'H:\\StuphMUD\\lib\\python', '-C', '.mud-config.cfg', '-w', 'H:\\StuphMUD\\lib\\world', '--cascade', '-z', 'index.mini', '--libdata=H:\\StuphMUD\\lib\\python\\supplemental\\iibdata', '-p', '9876', '--headless', '-vsa']

def runCore(options, args, globalize = True,
            consoleBaseClass = None,
            initForCore = None):

    if options.mud_package and isinstance(options.mud_package, str):
        from game import registerRealPath
        registerRealPath(options.mud_package)

    # I don't like this here, but it can't be a part of Core because
    # of the complication of the world module being separate from the
    # mud package.  So even though the event bridge could be initialized
    # as part of the mud package domain, the framework implementation
    # requires that it be part of the game module domain.
    from game import installBridge
    installBridge()

    # Todo: Stitch these two calls together within the compartment.
    buildCli(consoleBaseClass)

    from stuphos.stuphlib import wldlib, dblib
    from stuphos import stuphlib

    from game import world, Core

    return Core((options, args),
                stuphlib, world,
                globalize = globalize,
                initForCore = initForCore,
                consoleClass = Console)


class PythonWinCompartment:
    class PyWinConsoleBusiness: # XXX :skip:: (ConsoleBusiness): so much for a base class.
        # Usage: console.startInteractiveHandler()
        def __init_entity__(self, *args, **kwd):
            ConsoleBusiness.__init_entity__(self, *args, **kwd)
            self.readQ = queue.Queue()
            self.leavingMode = False
            self.debugMode = False

            # Todo: formalize this definition, not here locally:
            #    sh = Interactive.Shell.New()
            #    sh.handleCommand = self.handlePyWinInteractiveShellCommand
            #
            #    def handlePyWinInteractiveShellCommand(self, line):
            #        self.readQ.put(line)
            #        return True

            from common.subsystem.pythonwin.ide import Interactive
            class PyWinInteractiveShell(Interactive.Shell):
                def handleCommand(this, line):
                    # How to escape..?
                    # Or, system.core.interactive.pythonInterpreter .runcode
                    #
                    # Implement a top-level command-interpreter here for
                    # escaping (self.leaveGameMode)
                    #
                    # Actually, just make this routine return False..!

                    self.readQ.put(line)
                    self.leavingMode = False

                    if self.debugMode:
                        from pywin.debugger import set_trace
                        set_trace()

                    if 1:
                        print()
                        self.handleConsoleInput()

                    return not self.leavingMode

            self.pyWinInteractiveShellClass = PyWinInteractiveShell

        def textout(self, text):
            # When used as withPeerHead entity.
            out = sys.stdout
            if out is self:
                out = sys.interactive.interpreter

            return out.write(text) # should this be returning value..?

        def page_string(self, string):
            # todo: open document template
            print(string)

        def editString(self, string):
            if callable(self.messenger):
                self.messenger(self, string)

        def startInteractiveHandler(self):
            self.pyWinInteractiveShellClass(system.core.interactive)
        enterGameMode = startInteractiveHandler

        def leaveGameMode(self):
            self.leavingMode = True

        def readConsoleInput(self):
            return self.readQ.get()

        def perform(self, engine):
            if 0:
                # The classic way is to do this within the event loop,
                # but that isn't inline with the GUI thread, so, make
                # the handleCommand method call the event.
                if not self.readQ.empty():
                    self.engine.event += self.handleConsoleInput

    def __init__(self, options_file):
        self.loadOptionsFile(options_file)
        self.recompartmentalize()

    def loadOptionsFile(self, options_file):
        (self.argv, self.paths, self.game_module) = self.parseOptionsFile(options_file)
    def loadRunCore(self, game_module):
        return __import__(game_module, globals(), locals(), ['']).runCore

    def parseOptionsFile(self, filename):
        # Prepare (options, args) output from file.
        game_module = None
        argv = []
        paths = []
        for line in open(filename):
            line = line.strip()
            if line and line[0] in '#;':
                continue # Comments.

            colon = line.find(':')
            if colon > 0:
                option = line[:colon]
                value = line[colon + 1:].lstrip()

                if option in ('switches', 'switch'):
                    argv.append('-' + value)
                elif option == 'toggle':
                    argv.append('--' + value)
                elif option == 'system-path':
                    if value not in paths:
                        paths.append(value)
                elif option == 'game-module':
                    game_module = value
                elif len(option) == 1:
                    argv.append('-' + option)
                    argv.append(value)
                else:
                    argv.append('--' + option)
                    argv.append(value)

        return (argv, paths, game_module)

    def recompartmentalize(self):
        # Remove this directory from the system path.
        this_path = realpath(normpath(dirname(__file__)))
        if this_path in sys.path:
            sys.path.remove(this_path)

        # Add in new directories.
        for path in self.paths:
            if path not in sys.path:
                sys.path.append(path)

        # Load the game module runCore routine.
        self.runCore = self.loadRunCore(self.game_module)

        # Reparse the assembled command line.
        from game import parseCmdln
        (options, args) = parseCmdln(self.argv)

        # Don't let Core do async -- we'll handle that.
        options.asynchronous = False
        self.options = options
        self.args = args

    def Setup(self, core):
        #: Compartmentalize within the mud core:

        # Override the default system debug decorator.
        ##    from mud.tools import registerBuiltin
        ##    self.installPywinDebugger(registerBuiltin)

        # Register onShutdown event -- to quit application last!
        from stuphos import on
        @on.shutdownGame.last
        def exitApp():
            self.thisApp.OnQuit()

        # Install Console Class.
        ##    @runtime.api('Game::Support::Console')
        ##    @apply
        ##    class ConsoleAPI:
        ##        @breakOn
        ##        def create(self, *args, **kwd):
        ##            from game.cli import Console
        ##            return Console.create(*args, **kwd)

    def installPywinDebugger(self, registerBuiltin):
        # Declare the pywin-debug decorator.
        def pywinBreakOn(function):
            from pywin.debugger import runcall
            def pywinDebugCall(*args, **kwd):
                return self.runInGUIFor(runcall, function, *args, **kwd)

            return pywinDebugCall

        registerBuiltin(pywinBreakOn, 'breakOn')

    def _getCoreBootstrap(self):
        def resumeCoreBootstrap():
            # Debugging bootstrap:
            # installDefaultDebugger(defaultRegisterBuiltin)
            # self.installPywinDebugger(defaultRegisterBuiltin)

            from _thread import start_new_thread as nth
            nth(self.runCore, (self.options, self.args),
                dict(consoleBaseClass = self.PyWinConsoleBusiness,
                     initForCore = self.Setup))

        return resumeCoreBootstrap

    def Run(self):
        sys.argv = ['/new']
        from pywin.framework import intpyapp

        thisApp = self.thisApp = intpyapp.thisApp

        thisApp.InitInstance()
        self.runInGUI(self._getCoreBootstrap())
        thisApp.Run()

    def runInGUI(self, function, *args, **kwd):
        def guiCompartment(handler, count):
            self.thisApp.DeleteIdleHandler(guiCompartment)
            function(*args, **kwd)

        self.thisApp.AddIdleHandler(guiCompartment)

    def runInGUIFor(self, function, *args, **kwd):
        waitQ = queue.Queue()
        def guiCompartment(handler, count):
            self.thisApp.DeleteIdleHandler(guiCompartment)
            waitQ.put(function(*args, **kwd))

        self.thisApp.AddIdleHandler(guiCompartment)
        return waitQ.get()

    # Embed in already-running PythonWin.
    @classmethod
    def EmbedFromOptions(self, xxx_todo_changeme, asynchronous = True):
        (options, args) = xxx_todo_changeme
        kwd = dict(consoleBaseClass = self.PyWinConsoleBusiness)

        if asynchronous:
            from _thread import start_new_thread as nth
            return nth(runCore, (options, args), kwd)

        return runCore((options, args), **kwd)

    @classmethod
    def EmbedFromCmdln(self, cmdln, *args, **kwd):
        if isinstance(cmdln, str): # boxing
            from shlex import split as splitShlex
            cmdln = splitShlex(cmdln)

        from game import parseCmdln
        return self.EmbedFromOptions(parseCmdln(cmdln), *args, **kwd)

def PyWinEmbeddedMain():
    cmdln = pythonWin.property['Startup StuphMUD Command Line']

    if cmdln:
        PythonWinCompartment.EmbedFromCmdln(cmdln)


try: from common.runtime.structural.document import Submapping
except ImportError: pass
else:
    from common.runtime import Object

    class wmc:
        class Factory(Submapping):
            class wmcStuphOSConfiguration(Object):
                '''
                Westmetal Configuration::
                    stuphos = game.cli.wmc

                stuphos:
                    cmdln(stuphos$configuration):
                        game-configuration: '{support}(".mud.config")'
                        data-folder: '{support}.folder.lib'

                        # no-world: true
                        world-folder: '{support}.folder.lib.world'
                        load-world-cascade: true
                        zone-index: 'index.mini'

                        port: 6112
                        mud-package: '{support}.folder.lib.python'

                        interactive: false
                        asynchronous: true
                        debug: false
                        # verbose: 2

                        implementor: true
                        headless: false

                    init(python$module)::
                        builtin(bootStuph = container.cmdln)

                '''

                def __init__(self, name, value, **kwd):
                    self.gameConfiguration = value.get('game-configuration')
                    self.dataFolder = value.get('data-folder')

                    self.noWorld = value.get('no-world')
                    self.worldFolder = value.get('world-folder')
                    self.loadWorldCascade = value.get('load-world-cascade')
                    self.zoneIndex = value.get('zone-index')

                    self.port = value.get('port')
                    self.mudPackage = value.get('mud-package')

                    self.interactive = value.get('interactive')
                    self.asynchronous = value.get('asynchronous')
                    self.debug = value.get('debug')
                    self.verbose = value.get('verbose')

                    self.implementor = value.get('implementor')
                    self.headless = value.get('headless')

                def prepareCmdln(self):
                    # Todo: translate contained structural configuration
                    # to expected cli-based options.

                    ##    ('-w', '--world-dir')
                    ##    ('-z', '--zone-index', '--index')
                    ##    ('-i', '--interactive', action = 'store_true')
                    ##    ('-a', '--async', action = 'store_true')
                    ##    ('-W', '--cascade', '--load-world', action = 'store_true')
                    ##    ('-C', '--config-file', '--config', '--game-config')
                    ##    ('-g', '--debug', action = 'count', default = 0)
                    ##    ('-n', '--no-world', action = 'store_true')
                    ##    ('-p', '--port', type = int)
                    ##    ('-m', '--mud-package', '--mud')
                    ##    ('-s', '--supreme', action = 'store_true')
                    ##    ('-L', '--data-dir', '--lib-dir')
                    ##    ('--libdata')
                    ##    ('-v', '--verbose', action = 'count', default = 0)
                    ##    ('--headless', '--no-console', action = 'store_true')

                    return (dict(),
                            ())

                def embed(self, *args, **kwd):
                    (cmdlnOptions, cmdlnArgs) = self.prepareCmdln()
                    return PythonWinCompartment.EmbedFromOptions \
                           ((cmdlnOptions, cmdlnArgs),
                            *args, **kwd)

                __call__ = boot = embed

            configuration = wmcStuphOSConfiguration


##    def defaultRegisterBuiltin(function, name):
##        setattr(__builtins__, name, function)
##    def installDefaultDebugger(registerFunction):
##        from pdb import set_trace
##        def breakOn(function):
##            def debugCall(*args, **kwd):
##                set_trace()
##                return function(*args, **kwd)
##
##            return debugCall
##        registerFunction(breakOn, 'breakOn')

# Todo: When run from the support (development) directory,
# path will be relative.  Can we do this kind of thing sanely?
##    def qualifyModulePath(module):
##        p = module.__path__
##        p[0] = abspath(p[0])
##
##    qualifyModulePath(__import__('game'))
# And is it really going to be effective?

def main(argv = None):
    # Application compartment.
    from game import getCmdln, parseCmdln

    argv = getCmdln(argv)
    if len(argv) == 3 and argv[1] == '/pywin':
        # Host the MUD server asynchronously while running pywin in main.
        pywin = PythonWinCompartment(argv[2])
        pywin.Run()
    else:
        return runCore(*parseCmdln(argv))

if __name__ == '__main__':
    main()
else:
    buildCli()



# Configuration Object Model.

from stuphos.etc.tools.strings import isYesValue

# Paths.
from os.path import join as joinpath, normpath, dirname
import sys

try: from path import path as pathObject
except ImportError:
    pathObject = str

def getPlatformDrivePath(drive_letter):
    if sys.platform.startswith('cygwin'):
        return '/cygdrive/%s' % str(drive_letter).lower()
    if sys.platform.startswith('win32'):
        return '%s:\\' % str(drive_letter).upper()

    return '/%s' % str(drive_letter)

def getStandardDrivePath(drive_letter, *parts):
    return getStandardPath(getPlatformDrivePath(drive_letter), *parts)

def getStandardPath(*parts):
    return pathObject(normpath(joinpath(*parts)))

def getParentPath(path, level = 1):
    # Standardize?
    for x in range(level):
        path = dirname(path)

    return path

# Configuration Files.
from configparser import ConfigParser, DEFAULTSECT, NoOptionError, NoSectionError

def loadConfig(filename, defaults = None):
    config = ConfigParser(defaults = defaults)
    config.read([filename])
    return Configuration(config, filename)

def loadConfigFromString(string, defaults = None, name = None):
    config = ConfigParser(defaults = defaults)
    config.read_string(string, source = name)
    return Configuration(config, name)

class Configuration:
    class Section:
        def __init__(self, config, section = None, **vars):
            if section is not None and type(section) is not str:
                raise TypeError(type(section).__name__)

            self.config = config
            self.section = section or DEFAULTSECT
            self.vars = vars

        def get(self, name, default = None):
            try: return self.config._config.get(self.section, name, vars = self.vars)
            except (NoOptionError, NoSectionError): return default

        def options(self):
            return self.config.options(self.section)

        def __iter__(self):
            return iter(self.options())
        def __getitem__(self, name):
            return self.get(name) # ??? Should this throw error?
        def __repr__(self):
            return "<Section '%s:%s'>" % (self.config.filename, self.section)

        @property
        def data(self):
            return dict((n, self.get(n)) for n in self.options())

        def setOption(self, name, value):
            return self.config._config.set(self.section, name, value)


    class _addressor:
        # todo: rewrite this with attributable.
        class _section:
            def __init__(self, addressor, name):
                self._addressor = addressor
                self._section = name

            def __iter__(self):
                # debugOn()
                for name in self._addressor._config.options(self._section):
                    yield (name, getattr(self, name))

            def __getattr__(self, option):
                try: return object.__getattribute__(self, option)
                except AttributeError:
                    option = option.replace('_', '-')
                    return self._addressor.getSectionOption \
                        (self._section, option)

        def __init__(self, config):
            self._config = config

        def __getattr__(self, name):
            try: return object.__getattribute__(self, name)
            except AttributeError:
                return self._section(self, name)

        def getSectionOption(self, section, option):
            return self._config.get(option, section)


    class _addressorTruth(_addressor):
        def getSectionOption(self, section, option):
            return isYesValue(self._config.get(option, section))


    def __init__(self, config, filename = None):
        self._config = config
        self.filename = filename

    def get(self, name, section = None, vars = None):
        try: return self._config.get(section, name, vars = vars)
        except (NoOptionError, NoSectionError): pass

    def sections(self):
        return self._config.sections()
    def options(self, section = None):
        try: return self._config.options(section or DEFAULTSECT)
        except NoSectionError: return []

    def getSection(self, section = None, **vars):
        return self.Section(self, section, **vars)
    def getDefaultSection(self, **vars):
        return self.getSection(self, **vars)

    def __iter__(self):
        return iter(self.sections())
    def __getitem__(self, section):
        return self.getSection(section)
    def __repr__(self):
        return "<Configuration '%s'>" % self.filename

    @property
    def addressor(self):
        return self._addressor(self)

    @property
    def addressorTruth(self):
        return self._addressorTruth(self)


    def loadSetOptionList(self, options):
        return self.loadSetOption(self.parseSetOptionList(options))

    def loadSetOption(self, options):
        for (section, name, value) in options:
            if section and name and value:
                self.getSection(section).setOption(name, value)

    def parseSetOptionList(self, options):
        return map(self.parseSetOption, options)

    def parseSetOption(self, opt):
        try: (name, value) = opt.split('=', maxsplit = 1)
        except ValueError as e:
            raise ValueError(opt) from e

        try: (section, name) = name.split(':', maxsplit = 1)
        except ValueError as e:
            raise ValueError(name) from e

        yield section
        yield name
        yield value


# Package Support.
class PackageManager:
    def __init__(self, site_path = None, *components):
        if site_path is None:
            raise ValueError('Unsupported Site Path')

        self.site_path = site_path
        self.component_files = list(components)

    def install(self):
        from site import addpackage
        from sys import path as known_paths

        known_paths = set(known_paths)
        for comp_fl in self.component_files:
            # print(f'[addpackage] {self.site_path} / {comp_fl}')
            addpackage(self.site_path, comp_fl, known_paths)

# Standard MUD Properties.
LIB_ETC = getStandardPath('etc')



# Tool for enabling breakpoints on django/urlconf views.
from django.core.urlresolvers import get_resolver, get_urlconf, RegexURLResolver, RegexURLPattern

def findUrlPattern(path, resolver = None):
    if resolver is None:
        resolver = get_resolver(get_urlconf())

    match = resolver.regex.search(path)
    if match:
        if isinstance(resolver, RegexURLPattern):
            return resolver

        assert isinstance(resolver, RegexURLResolver)
        path = path[match.end():]
        for pattern in resolver.url_patterns:
            result = findUrlPattern(path, pattern)
            if result is not None:
                return result

_url_breakpoints = []

def doBreakOnUrl(peer, action = None, url = None):
    if action in ('show', None):
        peer.page_string('\r\n'.join(map(str, _url_breakpoints)) + '\r\n')
    else:
        # Enable breakpoint on django urlconf callback.
        pattern = findUrlPattern(url)
        if not pattern:
            print('Url not found: %r' % url, file=peer)

        elif action == 'enable':
            if enableUrlBreakpoint(pattern):
                print('Breaking on: %r' % pattern, file=peer)

        elif action == 'disable':
            if disableUrlBreakpoint(pattern):
                print('No longer breaking on: %r' % pattern, file=peer)

def enableUrlBreakpoint(pattern):
    if not hasattr(pattern, '__unbroken_callback'):
        unbroken = pattern._callback
        pattern._callback = breakOn(unbroken)
        setattr(pattern, '__unbroken_callback', unbroken)

        _url_breakpoints.append(pattern)
        return True

def disableUrlBreakpoint(pattern):
    try: unbroken = getattr(pattern, '__unbroken_callback')
    except AttributeError: pass
    else:
        pattern._callback = unbroken
        delattr(pattern, '__unbroken_callback')

        _url_breakpoints.remove(pattern)
        return True



from . import ASPELL

def spell(name, value, **kwd):
	@ASPELL(name, value['max_mana'], value['min_mana'], value['mana_change'],
			value.get('minpos', 'Standing'), value['targets'], value['violent'],
			value.get('wearoff', ''), register = False)

	def girlSpell(spell, caster):
		# todo: programmer
		task = value['manual'](environ = dict(),
							   locals = dict(environment = kwd['container'],
											 caster = caster,
											 victim = None))

		# task.name = '%s' % spell

	return girlSpell



# Runtime Error Handling.
__all__ = ['HandleException', 'ShowFrame',
           'getSystemException', 'getSystemExceptionString',
           'reraiseSystemException', 'reraise',
           'getModuleFileBasename']

from traceback import format_exc
from sys import exc_info as getSystemException

from . import getline, basename

# Utilities.
def getSystemExceptionString():
    return str(format_exc())
def reraiseSystemException():
    reraise(*getSystemException())
def reraise(type, value, tb):
    # I know this is the traditional pattern with python 3 traceback raising,
    # but usually value is already an instance of type.
    # raise type(value).with_traceback(tb)
    raise value.with_traceback(tb)

def getModuleFileBasename(filename):
    from sys import path as system_path
    for p in system_path:
        if p:
            if filename.startswith(p):
                return filename[len(p)+1:]

    return filename

# Game-Level Routines.
# Todo: rewrite this because obviously it's doing what .logs does.
def HandleException(exc = None):
    from .logs import log

    if exc is None:
        exc = getSystemException()

    (tp, val, tb) = exc

    # Find (second to?) last frame.
    while tb.tb_next:
        tb = tb.tb_next

    frame = tb.tb_frame
    code  = frame.f_code

    log('%s(%s): %s (%s:%d)' % (getattr(tp, '__name__', '<Unnamed>'),
                                code.co_name, str(val),
                                code.co_filename, frame.f_lineno))

def ShowFrame(frame, name, exc = None, use_basename = True):
    code = frame.f_code
    filename = code.co_filename

    line = getline(filename, frame.f_lineno).strip()

    if use_basename == 'relative':
        filename = system.relative(filename) # WRLC
    elif use_basename:
        filename = basename(filename)

    if exc:
        return '%s in %s: %s\n -> (%s:%d) %s' % \
               (name, code.co_name, exc[1], filename, frame.f_lineno, line)

    return '    (%s:%d:%s) %s' % (filename, frame.f_lineno, code.co_name, line)


class CodeSourceServer(dict):
    _lineClassCache = dict()

    @property
    def patterns(self):
        for (pattern, source) in self.items():
            if isinstance(pattern, str):
                pattern = re.compile(pattern)

            yield (pattern, source)

    def locate(self, code):
        filename = code.co_filename
        for (pattern, sourceClass) in self.patterns:
            m = pattern.match(filename)
            if m is not None:
                return sourceClass(self, code, *m.groups(),
                                   **m.groupdict())

    __getitem__ = __call__ = locate

    def sourceCodeServer(self, reader):
        return newClassObject('$', (self.CodeSource,), dict(read = reader))

    class CodeSource:
        def __init__(self, server, code, path):
            self.server = server
            self.code = code
            self.path = path

        @property
        def cacheId(self):
            cls = classOf(self)
            return '%s:%s' % (moduleOf(cls), nameOf(cls))

        @property
        def lineCache(self):
            id = self.cacheId
            try: return self.server._lineClassCache[id]
            except KeyError:
                cache = self.server._lineClassCache[id] = dict()
                return cache

        def __getitem__(self, index):
            cache = self.lineCache
            try: lines = cache[self.path]
            except KeyError:
                lines = cache[self.path] = self.read()

            return lines.__getitem__(index)

        def read(self):
            raise NotImplementedError

        @property
        def content(self):
            return self.read()
        source = content


class Traceback:
    def __init__(self, ptr):
        self.ptr = ptr

    @classmethod
    def StackFrom(self, tb):
        return self(tb).stack

    @classmethod
    def InstallCodeServer(self, pattern, server):
        self.Frame.CodeSource.server[pattern] = server.sourceCodeServer(server)

    @classmethod
    def Bind(self, pattern):
        def bind(server):
            self.InstallCodeServer(pattern, server)

        return bind

    @property
    def previous(self):
        return classOf(self)(self.ptr.tb_back)
    back = previous

    def __iter__(self):
        this = self
        while this is not None:
            yield this
            this = this.previous

    @property
    def frame(self):
        return self.Frame(self, self.ptr.tb_frame)

    @property
    def stack(self):
        return self.Stack(self)

    class Stack:
        def __init__(self, head):
            self.head = head

        def __iter__(self):
            return iter(reversed(list(self.head)))

        @property
        def rendering(self):
            return '\n'.join(f.frame.rendering for f in self)


    class Frame:
        def __init__(self, tb, ptr):
            self.tb = tb
            self.ptr = ptr

        @property
        def codeSource(self):
            return self.CodeSource(self, self.ptr.f_code)
        source = codeSource

        @property
        def lineNumber(self):
            return self.ptr.f_lineno

        @property
        def rendering(self):
            code = self.source
            lnr = self.lineNumber

            path = system.relative(code.path)
            r = '[%s:%s]\n' % (path, lnr)

            try: source = self.source[lnr]
            except: pass
            else:
                r += indent(source)

            return r

        def __str__(self):
            return self.rendering


        class CodeSource:
            server = CodeSourceServer()

            def __init__(self, f, code):
                self.f = f
                self.code = code

            @property
            def fileName(self):
                return self.code.co_filename

            path = fileName

            @property
            def sourceCode(self):
                server = self.server.locate(self.code)
                return server.source # .replace('\r', '')
            source = sourceCode

            @property
            def sourceSequence(self):
                return self.sourceCode.split('\n')

            def __iter__(self):
                return iter(self.sourceSequence)

traceback = Traceback.StackFrom
sourceClass = CodeSourceServer.CodeSource
codeServer = Traceback.InstallCodeServer
bindCodeServer = Traceback.Bind


# @bindCodeServer('^/virtual/(?P<path>.*)$')
# def virtualSource(code):
#     core = runtime.Agent.System(None)
#     node = core[code.path]

#     if isinstance(node, core.Node.Document):
#         return node.content
